//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#ifndef MSL_GEOMPQP_H
#define MSL_GEOMPQP_H

#include <LEDA/file.h>
#include <LEDA/array.h>
#include <LEDA/array2.h>
#include <LEDA/list.h>
#include <LEDA/matrix.h>
#include <LEDA/point.h>
#include <LEDA/polygon.h>
#include <LEDA/stream.h>
#include <LEDA/string.h>
#include <LEDA/vector.h>
#include <LEDA/graph.h>
#include <LEDA/d3_point.h> 

#include "geom.h"
#include "3Dtriangle.h"

#include <PQP/PQP.h>

#include <LEDA/REDEFINE_NAMES.h>

// OK, this is bad.  The maximum number should be dynamic, but PQP
// really seems to want a static matrix.
#define MAXBODIES 100

//! Parent class PQP-based list of Triangle models

class GeomPQP: public Geom {
 protected:
  PQP_REAL RR[3][3],RO[3][3];
  PQP_REAL TR[3],TO[3];  
 public:
  list<Triangle> Obst;
  list<Triangle> Robot;
  PQP_Model Ro, Ob;
  GeomPQP(string path);
  virtual ~GeomPQP() {};
  virtual void LoadEnvironment(string path);
  virtual void LoadRobot(string path);
  virtual bool CollisionFree(const vector &q){return 0;} 
  virtual double DistanceComp(const vector &q){return 0;}
  virtual list<polygon> EnvironmentToLedaPolygons();
  virtual list<polygon> RobotToLedaPolygons(const vector &q);
};


//! A parent class for 2D PQP geometries

class GeomPQP2D: public GeomPQP {
 public:
  list<polygon> ObstLedaPolygons;
  list<polygon> BorderLedaPolygons;
  list<polygon> RobotLedaPolygons;
  GeomPQP2D(string path);
  virtual ~GeomPQP2D() {};
  virtual void LoadEnvironment(string path);
  virtual void LoadRobot(string path);
  virtual list<polygon> EnvironmentToLedaPolygons();
};


//! 2D point robot

class GeomPQP2DPoint: public GeomPQP2D {
 public:
  GeomPQP2DPoint(string path);
  virtual ~GeomPQP2DPoint() {};
  virtual bool CollisionFree(const vector &q); // Input is configuration
  virtual double DistanceComp(const vector &q);  // Distance in world
};


//! 2D rigid body 

class GeomPQP2DRigid: public GeomPQP2D {
 public:
  GeomPQP2DRigid(string path);
  virtual ~GeomPQP2DRigid() {};
  virtual bool CollisionFree(const vector &q); // Input is configuration
  virtual double DistanceComp(const vector &q);  // Distance in world
  virtual vector ConfigurationDifference(const vector &q1, 
					      const vector &q2); 
  virtual list<polygon> RobotToLedaPolygons(const vector &q);
  void SetTransformation(const vector &q); // Input is configuration
};



//! A collection of 2D rigid bodies

class GeomPQP2DRigidMulti: public GeomPQP2DRigid {
 private:
  array<list<Triangle> > Robot;
  array<PQP_Model> Ro;
  list<vector> CollisionPairs; // Index pairs to check for collision
  PQP_REAL TR[MAXBODIES][3];  // Robot translation
  PQP_REAL RR[MAXBODIES][3][3];  // Robot rotation
 public:
  bool SelfCollisionCheck;
  GeomPQP2DRigidMulti(string path);
  virtual ~GeomPQP2DRigidMulti() {};
  virtual bool CollisionFree(const vector &q); // Input is configuration
  virtual double DistanceComp(const vector &q);  // Distance in world
  virtual void LoadRobot(string path); // Load multiple robots
  void SetTransformation(const vector &q); // Input is configuration
};



//! 3D rigid body

class GeomPQP3DRigid: public GeomPQP {
 public:
  GeomPQP3DRigid(string path);
  virtual ~GeomPQP3DRigid() {};
  virtual bool CollisionFree(const vector &q); // Input is configuration
  virtual double DistanceComp(const vector &q);  // Distance in world
  virtual vector ConfigurationDifference(const vector &q1, 
					      const vector &q2); 
  void SetTransformation(const vector &q); // Input is configuration
};


//! A collection of 3D rigid modies
class GeomPQP3DRigidMulti: public GeomPQP3DRigid {
 private:
  array<list<Triangle> > Robot;
  array<PQP_Model> Ro;
  list<vector> CollisionPairs; // Index pairs to check for collision
  PQP_REAL TR[MAXBODIES][3];  // Robot translation
  PQP_REAL RR[MAXBODIES][3][3];  // Robot rotation
 public:
  bool SelfCollisionCheck;
  GeomPQP3DRigidMulti(string path);
  virtual ~GeomPQP3DRigidMulti() {};
  virtual bool CollisionFree(const vector &q); // Input is configuration
  virtual double DistanceComp(const vector &q);  // Distance in world
  virtual void LoadRobot(string path); // Load multiple robots
  void SetTransformation(const vector &q); // Input is configuration
};

#include <LEDA/UNDEFINE_NAMES.h>

#endif






