#include <math.h>
#include <stdlib.h>

#include <LEDA/window.h>

#include <MSL/rrt.h>
#include <MSL/model2d.h>

#include <LEDA/REDEFINE_NAMES.h>



void DrawPath(const list<vector> &path, Planner *Pl, window *w)
{
  vector x1,x2;
  double tranx,trany,scalex,scaley;
  double lw;

  lw = w->get_line_width();
  w->set_line_width(5.0);

  // The default window coords are [0,100]x[0,100]
  scalex = 100.0/(Pl->P->UpperState[0] - 
		  Pl->P->LowerState[0]); 
  scaley = 100.0/(Pl->P->UpperState[1] - 
		  Pl->P->LowerState[1]);
  tranx = -1.0 * scalex * Pl->P->LowerState[0];
  trany = -1.0 * scaley * Pl->P->LowerState[1];

  x1 = path.head();
  forall(x2,path) {
    w->draw_segment(x1[0]*scalex+tranx,
		    x1[1]*scaley+trany,
		    x2[0]*scalex+tranx,
		    x2[1]*scaley+trany,violet);
    x1 = x2;
  }

  w->set_line_width(lw);
}




void DrawRRT(Planner *Pl, window *w)
{
  edge e;
  vector x1,x2;
  double tranx,trany,scalex,scaley;
  
  // The default window coords are [0,100]x[0,100]
  scalex = 100.0/(Pl->P->UpperState[0] - 
		  Pl->P->LowerState[0]); 
  scaley = 100.0/(Pl->P->UpperState[1] - 
		  Pl->P->LowerState[1]); 
  tranx = -1.0 * scalex * Pl->P->LowerState[0];
  trany = -1.0 * scaley * Pl->P->LowerState[1];

  // Show first tree
  if (Pl->G.number_of_nodes() > 0) {
    forall_edges(e,Pl->G) {
      x1 = Pl->G.inf(Pl->G.source(e));
      x2 = Pl->G.inf(Pl->G.target(e));
      w->draw_segment(x1[0]*scalex+tranx,
		      x1[1]*scaley+trany,
		      x2[0]*scalex+tranx,
		      x2[1]*scaley+trany,blue);
    }
  }
  // Show second tree (if it exists)
  if (Pl->G2.number_of_nodes() > 0) {
    forall_edges(e,Pl->G2) {
      x1 = Pl->G2.inf(Pl->G2.source(e));
      x2 = Pl->G2.inf(Pl->G2.target(e));
      w->draw_segment(x1[0]*scalex+tranx,x1[1]*scaley+trany,
		      x2[0]*scalex+tranx,x2[1]*scaley+trany,
		      red);
    }
  }
}




int main(int argc, char *argv[])
{
  string path = "";
  window *w;
  Model *m = NULL;
  Geom *g = NULL;
  Problem *prob;
  RRT *rrt;
  int i;
  list<vector> solution;

  // Initialize the window
  w = new window(600,600,"Rapidly-Exploring Random Trees");
  w->set_line_width(2.0);
  w->display();
  w->draw_rectangle(0.0,0.0,100.0,100.0,red);


  leda_wait(2.0);

  cout << "Draw an RRT starting in the center of a box\n";
  m = new Model2DPoint(path); // Define the differential model
  g = new GeomLedaPolygonal(path); 
  prob = new Problem(g,m,path); // This is the input to a planner
  prob->InitialState = vector(50.0,50.0);
  prob->GoalState = vector(100.0,100.0);
  rrt = new RRT(prob);
  rrt->NumNodes = 50;
  rrt->PlannerDeltaT = 1.0;

  for (i = 0; i < 15; i++) {
    rrt->Construct();
    DrawRRT(rrt,w);
    leda_wait(0.2);
  }
  for (i = 0; i < 30; i++) {
    rrt->Construct();
    DrawRRT(rrt,w);
    leda_wait(0.05);
  }

  leda_wait(2.0);

  cout << "Draw an RRT starting in the corner\n";
  w->clear();
  w->draw_rectangle(0.0,0.0,100.0,100.0,red);
  rrt->Reset();
  rrt->NumNodes = 50;
  rrt->PlannerDeltaT = 1.0;
  prob->InitialState = vector(1.0,1.0);
  for (i = 0; i < 15; i++) {
    rrt->Construct();
    DrawRRT(rrt,w);
    leda_wait(0.1);
  }
  for (i = 0; i < 30; i++) {
    rrt->Construct();
    DrawRRT(rrt,w);
    leda_wait(0.05);
  }

  leda_wait(2.0);

  cout << "Draw an RRT for the car\n";
  w->clear();
  w->draw_rectangle(0.0,0.0,100.0,100.0,red);
  rrt->Reset();
  rrt->NumNodes = 50;
  rrt->PlannerDeltaT = 1.0;
  m = new Model2DRigidCar(path); // Define the differential model
  prob->SetModel(m);
  prob->InitialState = vector(50.0,50.0,0.0);
  prob->GoalState = vector(100.0,50.0,0.0);
  for (i = 0; i < 50; i++) {
    rrt->Construct();
    DrawRRT(rrt,w);
    leda_wait(0.05);
  }

  leda_wait(2.0);

  cout << "Draw an RRT for the forward-only car\n";
  w->clear();
  w->draw_rectangle(0.0,0.0,100.0,100.0,red);
  rrt->Reset();
  rrt->NumNodes = 50;
  rrt->PlannerDeltaT = 1.0;
  m = new Model2DRigidCarForward(path); // Define the differential model
  prob->SetModel(m);
  prob->InitialState = vector(50.0,50.0,1.732);
  prob->GoalState = vector(100.0,50.0,1.732);
  for (i = 0; i < 50; i++) {
    rrt->Construct();
    DrawRRT(rrt,w);
    leda_wait(0.05);
  }

  // Finished
  cout << "Click a mouse button to terminate\n";
  w->read_mouse();
  return 0;
}


#include <LEDA/UNDEFINE_NAMES.h>
