//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#include <math.h>
#include <stdio.h>

#include "fdp.h"
#include "defs.h"

// This is needed because g++ does not support export keyword!
#include "marray.C"

#include <LEDA/REDEFINE_NAMES.h>

#define UNVISITED 0
#define VISITED 1   // Visited by the first tree, G
#define COLLISION 2
#define VISITED2 3  // Visited by the second tree, G2

// *********************************************************************
// *********************************************************************
// CLASS:     FDP base class
// 
// *********************************************************************
// *********************************************************************

FDP::FDP(Problem *problem): Planner(problem) {

  NumNodes = 20000;
  GridDefaultResolution = 50;
  Reset();
}



void FDP::Reset() {
  int i;
  bool done,carry;

  Planner::Reset();

  SatisfiedCount = 0;

  GEdgeTime.init(G,PlannerDeltaT);

  GridDimensions = array<int>(P->StateDim);
  GridDimensions.init(GridDefaultResolution);
  if (is_file(P->FilePath + "GridDimensions")) {
    file_istream fin(P->FilePath + "GridDimensions");
    fin >> GridDimensions;
    fin.close();
  }
  
  Grid = MultiArray<int>(GridDimensions, 0);

  Quantization = vector(P->StateDim);

  for (i = 0; i < P->StateDim; i++) {
    Quantization[i] = (P->UpperState[i] - P->LowerState[i])/GridDimensions[i];
  }

  if (!is_file(P->FilePath+"GapError")) {
    for (i = 0; i < P->StateDim; i++)
      GapError[i] = Quantization[i] / 2.0;
  }

  // Mark the collision region
  array<int> indices(P->StateDim);
  array<int> maxindices(P->StateDim);
  indices.init(0);
  for (i = 0; i < P->StateDim; i++)
    maxindices[i] = GridDimensions[i] - 1;
  
  // Loop through all of the indices and check each for collision
  cout << "Performing collision detection to initialize grid.\n";
  done = false;
  while (!done) {
    // Check for collision
    Grid[indices] = 
      (P->Satisfied(IndicesToState(indices))) ? UNVISITED : COLLISION;

    // Move to the next grid point
    if (indices[0] < GridDimensions[0] - 1)
      indices[0]++;
    else { // Carry
      indices[0] = 0;
      carry = true;
      i = 1;
      while ((carry)&&(i < P->StateDim)) {
	if (indices[i] < GridDimensions[i] - 1) {
	  indices[i]++;
	  carry = false;
	}
	else {
	  indices[i] = 0;
	  if (i == P->StateDim - 1)
	    done = true;
	}
	i++;
      }
    }
  }
}



void FDP::Construct()
{
  cout << "No effect -- Use Plan\n";
}


bool FDP::Plan()
{
  int i;
  node n,nn;
  vector nx,x,u;
  double ptime;
  list<node> path;
  pq_item pqi;
  double cost;
  array<int> indices;
  
  // Make the root node of G
  if (G.number_of_nodes() == 0) {
    n = G.new_node(P->InitialState);
    GEdgeTime[n] = 0.0;
    Q.clear();
    Q.insert(0.0,n);  // Add the first node to the queue
  }

  i = 0;
  while ((i < NumNodes)&&
	 (!Q.empty())) {

    // Remove the element with smallest cost
    pqi = Q.find_min();
    n = Q[pqi];
    cost = Q.prio(pqi);
    Q.del_item(pqi);
    x = G.inf(n);

    // Try all inputs
    forall(u,P->GetInputs(x)) {
      nx = P->Integrate(x,u,PlannerDeltaT);
      indices = StateToIndices(nx);
      // If we are visiting an UNVISITED place...
      if (Grid[indices] == UNVISITED) {
	Grid[indices] = VISITED;
	// Make a new node and edge
	nn = G.new_node(nx); // Make a new node with state nx
	G.new_edge(n,nn,u);
	Q.insert(SearchCost(cost,n,nn),nn); // Put it into the priority queue

	// Check if goal reached
	if (GapSatisfied(G.inf(nn),P->GoalState)) { 
	  cout << "Successful Path Found\n";
	  path = PathToLeaf(nn,G);
	  // Make the correct times
	  ptime = 0.0; TimeList.clear();
	  forall(n,path) {
	    ptime += GEdgeTime[n];
	    TimeList.push_back(ptime);
	  }

	  RecordSolution(path); // Write to Path and Policy
	  return true;
	}
      }
    }

    i++;
  }

  cout << "Failure to find a path\n";
  return false;
}


double FDP::SearchCost(double initcost, const node &n, const node &nn) {
  return initcost + PlannerDeltaT;
}


// The algorithm actually starts at the leaf and stops at the root
list<node> FDP::PathToLeaf(const node &n, const GRAPH<vector,vector> &g) 
{
   list<node> path;
   node ni;
   int i;

   path.clear();
   path.push(n); 

   i = 0;
   ni = n;
   while ((ni != g.first_node())&&(i < g.number_of_nodes())) {
     path.push(g.source(g.first_in_edge(ni)));
     ni = g.source(g.first_in_edge(ni));
     i++;
   }

   return path;
}


array<int> FDP::StateToIndices(const vector &x) {
  int i;

  array<int> indices(P->StateDim);

  for (i = 0; i < P->StateDim; i++) {
    indices[i] = (int) ((x[i] - P->LowerState[i])/Quantization[i]);
    // These are needed because the state is sometimes out of bounds.
    // The other way to fix this is to ensure that every instance of
    // Integrate in Model and subclasses stays within bounds.
    if (indices[i] < 0)
      indices[i] = 0;
    if (indices[i] >= GridDimensions[i])
      indices[i] = GridDimensions[i] - 1;
  }

  return indices;
}


vector FDP::IndicesToState(const array<int> &indices) {
  int i;

  vector x(P->StateDim);
  for (i = 0; i < P->StateDim; i++) {
    x[i] = Quantization[i]*(indices[i]+0.5) + P->LowerState[i];
  }

  return x;
}


// *********************************************************************
// *********************************************************************
// CLASS:     FDPStar
// 
// *********************************************************************
// *********************************************************************

FDPStar::FDPStar(Problem *problem): FDP(problem) {
}


// The cost always represents the cost to come using the Metric
// for each step, plus the cost to go from nn to the goal.
double FDPStar::SearchCost(double initcost, const node &n, const node &nn) {
  return initcost - 
    P->Metric(G.inf(n),P->GoalState) +
    P->Metric(G.inf(n),G.inf(nn)) + 
    P->Metric(G.inf(nn),P->GoalState);
}



// *********************************************************************
// *********************************************************************
// CLASS:     FDPBestFirst
// 
// *********************************************************************
// *********************************************************************

FDPBestFirst::FDPBestFirst(Problem *problem): FDP(problem) {
}


double FDPBestFirst::SearchCost(double initcost, const node &n, 
				const node &nn) {
  return P->Metric(G.inf(n),P->GoalState);
}





// *********************************************************************
// *********************************************************************
// CLASS:     FDPBi
// 
// *********************************************************************
// *********************************************************************

FDPBi::FDPBi(Problem *problem): FDP(problem) {

  G2EdgeTime.init(G2,PlannerDeltaT);
}



void FDPBi::Reset() {
  FDP::Reset();
  G2EdgeTime.init(G2,PlannerDeltaT);
}



bool FDPBi::Plan()
{
  int i,k;
  node n,nn,nn2;
  vector nx,x,u;
  double ptime;
  list<node> path;
  pq_item pqi;
  double cost;
  array<int> indices,indices2;
  bool match;

  // Make the root node of G
  if (G.number_of_nodes() == 0) {
    n = G.new_node(P->InitialState);
    GEdgeTime[n] = 0.0;
    Q.clear();
    Q.insert(0.0,n);  // Add the first node to the queue
  }

  // Make the root node of G2
  if (G2.number_of_nodes() == 0) {
    n = G2.new_node(P->GoalState);
    G2EdgeTime[n] = 0.0;
    Q2.clear();
    Q2.insert(0.0,n);  // Add the first node to the queue
  }

  nn2 = G.first_node(); // Just to keep warnings away

  i = 0;
  while ((i < NumNodes)&&
	 (!Q.empty())&&
	 (!Q2.empty())) {

    // ******** Handle the tree from the initial state *************
    // Remove the element with smallest cost
    pqi = Q.find_min();
    n = Q[pqi];
    cost = Q.prio(pqi);
    Q.del_item(pqi);
    x = G.inf(n);

    // Try all inputs
    forall(u,P->GetInputs(x)) {
      nx = P->Integrate(x,u,PlannerDeltaT);
      indices = StateToIndices(nx);

      // If we are visiting a place visited by G2...
      if (Grid[indices] == VISITED2) {
	// Make a new node and edge
	nn = G.new_node(nx); // Make a new node with state nx
	G.new_edge(n,nn,u);
	// Get the node in G2 that was visited
	forall_nodes(n,G2) {
	  indices2 = StateToIndices(G2.inf(n));
	  match = true;
	  for (k = 0; k < P->StateDim; k++)
	    if (indices[k] != indices2[k])
	      match = false;
	  if (match)
	    nn2 = n;
	}
	RecoverSolution(nn,nn2);
	cout << "Successful Path Found\n";
	return true;
      }

      // If we are visiting an UNVISITED place...
      if (Grid[indices] == UNVISITED) {
	Grid[indices] = VISITED;
	// Make a new node and edge
	nn = G.new_node(nx); // Make a new node with state nx
	G.new_edge(n,nn,u);
	Q.insert(SearchCost(cost,n,nn),nn); // Put it into the priority queue

	// Check if goal reached
	if (GapSatisfied(G.inf(nn),P->GoalState)) { 
	  cout << "Successful Path Found\n";
	  path = PathToLeaf(nn,G);
	  // Make the correct times
	  ptime = 0.0; TimeList.clear();
	  forall(n,path) {
	    ptime += GEdgeTime[n];
	    TimeList.push_back(ptime);
	  }

	  RecordSolution(path); // Write to Path and Policy
	  return true;
	}
      }
    }

    // ******** Handle the tree from the goal state *************
    // Remove the element with smallest cost
    pqi = Q2.find_min();
    n = Q2[pqi];
    cost = Q2.prio(pqi);
    Q2.del_item(pqi);
    x = G2.inf(n);

    // Try all inputs
    forall(u,P->GetInputs(x)) {
      nx = P->Integrate(x,u,-PlannerDeltaT);  // Reverse time integration
      indices = StateToIndices(nx);

      // If we are visiting a place visited by G...
      if (Grid[indices] == VISITED) {
	// Make a new node and edge
	nn = G2.new_node(nx); // Make a new node with state nx
	G2.new_edge(n,nn,u);
	// Get the node in G2 that was visited
	forall_nodes(n,G) {
	  indices2 = StateToIndices(G.inf(n));
	  match = true;
	  for (k = 0; k < P->StateDim; k++)
	    if (indices[k] != indices2[k])
	      match = false;
	  if (match)
	    nn2 = n;
	}
	RecoverSolution(nn2,nn);
	cout << "Successful Path Found\n";
	return true;
      }

      // If we are visiting an UNVISITED place...
      if (Grid[indices] == UNVISITED) {
	Grid[indices] = VISITED2;
	// Make a new node and edge
	nn = G2.new_node(nx); // Make a new node with state nx
	G2.new_edge(n,nn,u);
	Q2.insert(SearchCost(cost,n,nn),nn); // Put it into the priority queue

	// Check if initial reached
	if (GapSatisfied(G2.inf(nn),P->InitialState)) { 
	  cout << "Successful Path Found\n";
	  path = PathToLeaf(nn,G2); path.reverse();
	  // Make the correct times
	  ptime = 0.0; TimeList.clear();
	  forall(n,path) {
	    ptime += G2EdgeTime[n];
	    TimeList.push_back(ptime);
	  }

	  RecordSolution(path); // Write to Path and Policy
	  return true;
	}
      }
    }

    i++;
  }

  cout << "Failure to find a path\n";
  return false;
}


// This is duplicated from RRTDual
void FDPBi::RecoverSolution(const node &n1, const node &n2) {
  list<node> path,path2;
  double ptime;
  node n;

  path = PathToLeaf(n1,G);
  // Figure out forward timings
  ptime = 0.0; TimeList.clear();
  forall(n,path) {
    ptime += GEdgeTime[n];
    TimeList.push_back(ptime);
  }
  
  path2 = PathToLeaf(n2,G2); path2.reverse();
  // Figure out backwards timings
  ptime += PlannerDeltaT; // Add a time step for the gap
  forall(n,path2)  { // Compute total time in G2 part of path
    TimeList.push_back(ptime);
    ptime += G2EdgeTime[n];
  }
  
  //cout << "Path: " << path;
  //cout << "Path2: " << path2;
  //cout << "\n TimeList: " << TimeList << " L: " << TimeList.length();

  RecordSolution(path,path2); // Write to Path and Policy
}




#include <LEDA/UNDEFINE_NAMES.h>






