//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#ifndef MSL_FDP_H
#define MSL_FDP_H

#include <LEDA/file.h>
#include <LEDA/graph.h>
#include <LEDA/list.h>
#include <LEDA/polygon.h>
#include <LEDA/vector.h>
#include <LEDA/p_queue.h>

#include "marray.h"
#include "planner.h"

#include <LEDA/REDEFINE_NAMES.h>

/*! Forward Dynamic Programming, as proposed by Barraquand, Latombe, 
Algorithmica 10:6, pp. 121-155, 1993.  The solutions should be optimized 
with respect to time, although problems can be caused by quantization 
errors.

When an instance is constructed, a grid is initialized and each
element of checked for collision.  The test point for collision is the
center of the cell.  The current version does not use distance
computations; therefore, a large value for PlannerDeltaT might cause
collisions to be missed.

Make sure that PlannerDeltaT is large enough to allow the state to
move from one cell to another without in a single step.  For example,
if the quantization leads to a grid boundary every 2.0 units, then
PlannerDeltaT could be set to cause the state to change by 3.0 units.

GridDimensions sets the resolution of the grid and can be read from a file.
For high-dimensional problems an error message may occur due to a grid
that is too large.  To enable larger grids, set the MaxSize to a desirable
size in the MultiArray class (in marray.C).  */

//! A dynamic programming approach to nonholonomic planning, as proposed by Barraquand, Latombe, 
//! Algorithmica 10:6, pp. 121-155, 1993.

class FDP: public Planner {
 protected:

  //! Keep track of timings
  node_map<double> GEdgeTime;

  //! Priority queue of nodes
  p_queue<double,node> Q;

  //! A quantized state space
  MultiArray<int> Grid;

  //! Dimensions for the grid
  array<int> GridDimensions;

  //! Default size for each axis of the grid
  int GridDefaultResolution;

  //! The quantized step size for each axis (computed automatically)
  vector Quantization;

  virtual double SearchCost(double initcost, const node &n, const node &nn);

  //! Pull the path out of the tree
  list<node> PathToLeaf(const node &n, const GRAPH<vector,vector> &g);

  array<int> StateToIndices(const vector &x);

  vector IndicesToState(const array<int> &indices);

  public:

  //! A constructor that initializes data members.
  FDP(Problem *problem);

  //! Empty destructor
  ~FDP() {};
  
  //! Number of times the collision checker has been called
  int SatisfiedCount;

  //! Reset the planner
  virtual void Reset();

  //! Generate an FDP
  virtual void Construct();

  //! Attempt to solve an Initial-Goal query by growing an FDP
  virtual bool Plan();

};


//! An A-Star search variant. The Metric in Problem is used as the cost.
class FDPStar: public FDP {
 protected:
  virtual double SearchCost(double initcost, const node &n, const node &nn);
 public:
  FDPStar(Problem *p);
};
 

//! Best first search variant, using the Metric in Problem.
class FDPBestFirst: public FDP {
 protected:
  virtual double SearchCost(double initcost, const node &n, const node &nn);
 public:
  FDPBestFirst(Problem *p);
};


//! A bidirectional version of forward dynamic programming
class FDPBi: public FDP {
 protected:

  //! Keep track of timings
  node_map<double> G2EdgeTime;

  //! Priority queue of nodes
  p_queue<double,node> Q2;

  //! Pull out the path and timings from the graphs
  void RecoverSolution(const node &n1, const node &n2);

 public:

  //! A constructor that initializes data members.
  FDPBi(Problem *problem);

  //! Empty destructor
  ~FDPBi() {};

  //! Reset the planner
  virtual void Reset();

  //! Attempt to solve an Initial-Goal query by growing an FDP
  virtual bool Plan();
};


#include <LEDA/UNDEFINE_NAMES.h>

#endif


