//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#include <fstream.h>
#include <math.h>

#include "geom.h"
#include "defs.h"

#include <LEDA/REDEFINE_NAMES.h>

Geom::Geom(string path = "") {

  if ((path.length() > 0)&&(path[path.length()-1] != '/'))
    path += "/";
  FilePath = path;

  GeomDim = 3;
  if (is_file(FilePath + "GeomDim")) {
    file_istream fin(FilePath + "GeomDim");
    fin >> GeomDim;
  }
  else {
    cout << "Assuming default geometry dimension: " << GeomDim << "\n";
  }
}



vector Geom::ConfigurationDifference(const vector &q1, 
				     const vector &q2)
{
  return (q2 - q1);  // By default, assume no topological problems
}


// *********************************************************************
// *********************************************************************
// CLASS:     GeomLedaPolygonal
// 
// *********************************************************************
// *********************************************************************

// Constructor
GeomLedaPolygonal::GeomLedaPolygonal(string path = ""):Geom(path) {

  NumBodies = 1;
  
  if (is_file(FilePath + "Obst")) {
    file_istream fin(FilePath + "Obst");
    fin >> Obst;
  }

  MaxDeviates = vector(1.0,1.0);  // Translation only
}



// This simple collision checker uses LEDA's p.inside method.
// Certainly more efficient methods exist (incremental).
bool GeomLedaPolygonal::CollisionFree(const vector &x) {
  polygon p;

  forall(p,Obst) {
    if (p.inside(point(x))) {
      return false;
    }
  }

  return true;
}


double GeomLedaPolygonal::DistanceComp(const vector &x) {
  polygon p;
  double d_min,d;

  d_min = INFINITY;

  if (CollisionFree(x)) {
    forall(p,Obst) {
      d = p.distance(point(x));
      if (d < d_min)
	d_min = d;
    }
  }
  else
    d_min = -1.0;
  
  return d_min;
}



list<polygon> GeomLedaPolygonal::EnvironmentToLedaPolygons() {
  return Obst;
}



list<polygon> GeomLedaPolygonal::RobotToLedaPolygons(const vector &x) {
  list<polygon> trobot;

  point p1(0.5,0.5);
  point p2(-0.5,0.5);
  point p3(-0.5,-0.5);
  point p4(0.5,-0.5);
  list<point> pl;
  pl.push(p4); pl.push(p3); pl.push(p2); pl.push(p1);
  polygon p(pl);
  trobot.push(p.translate(x));
  
  return trobot;
}

 

// *********************************************************************
// *********************************************************************
// CLASS:     GeomLedaPolygonalRigid
// 
// *********************************************************************
// *********************************************************************

// Constructor
GeomLedaPolygonalRigid::GeomLedaPolygonalRigid(string path = ""):GeomLedaPolygonal(path) {

  if ((path.length() > 0)&&(path[path.length()-1] != '/'))
    path += "/";

  FilePath = path;
  
  if (is_file(FilePath + "Robot")) {
    file_istream fin(FilePath + "Robot");
    fin >> Robot;
  }

  // Compute the maximum deviates
  double dmax = 0.0;
  point pt;
  polygon p;
  forall(p,Robot) {
    forall_vertices(pt,p) {
      if (pt.distance() > dmax)
	dmax = pt.distance();
    }
  }
  MaxDeviates = vector(1.0,1.0,dmax);

}


// This is a very naive collision checker
// This should certainly be replaced some day with a 
// a more efficient algorithm.  It would be best to add
// a 2D incremental distance algorithm, based on the
// ideas of Lin and Canny, and Mirtich.  One could also
// use bitmaps since this is only 2D.

bool GeomLedaPolygonalRigid::CollisionFree(const vector &x) {
  polygon op,rp;
  segment seg,seg2;

  list<polygon> trobot = RobotToLedaPolygons(x);

  forall(rp,trobot) {
    forall(op,Obst) {
      forall_segments(seg,rp) {
	forall_segments(seg2,op) {
	  if (seg.intersection(seg2)) {
	    return false;
	  }
	}
      }
    }
  }

  // Check for containment
  forall(rp,trobot) {
    forall(op,Obst) {
      if (op.contains(rp.vertices().head()))
	return false;
    }
  }

  return true;
}


// This should certainly be replaced some day with a 
// a more efficient algorithm.  It would be best to add
// a 2D incremental distance algorithm, based on the
// ideas of Lin and Canny, and Mirtich.

double GeomLedaPolygonalRigid::DistanceComp(const vector &x) {
  polygon rp,op;
  double d_min,d;
  point rpt;

  d_min = INFINITY;

  list<polygon> trobot = RobotToLedaPolygons(x);

  // This is obviously inefficient
  if (CollisionFree(x)) {
    forall(rp,trobot) {
      forall_vertices(rpt,rp) {
	forall(op,Obst) {
	  d = op.distance(rpt);
	  if (d < d_min)
	    d_min = d;
	}
      }
    }
  }
  else
    d_min = -1.0;
  
  return d_min;
}



list<polygon> GeomLedaPolygonalRigid::RobotToLedaPolygons(const vector &q) {
  list<polygon> trobot;
  polygon p,tp;

  forall(p,Robot) {
    tp = p.rotate(q[2]).translate(q[0],q[1]);
    trobot.push_front(tp);
  }

  return trobot;
}



vector GeomLedaPolygonalRigid::ConfigurationDifference(const vector &q1, 
					   const vector &q2)
{
  vector dq(3);

  dq[0] = q2[0] - q1[0];
  dq[1] = q2[1] - q1[1];

  if (fabs(q1[2]-q2[2]) < PI)
    dq[2] = q2[2] - q1[2];
  else {
    if (q1[2] < q2[2])
      dq[2] = -(2.0*PI - fabs(q1[2]-q2[2]));
    else
      dq[2] = (2.0*PI - fabs(q1[2]-q2[2]));
  }

  return dq;
}

#include <LEDA/UNDEFINE_NAMES.h>
