//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#ifndef MSL_GEOM_H
#define MSL_GEOM_H

#include <LEDA/file.h>
#include <LEDA/array.h>
#include <LEDA/list.h>
#include <LEDA/matrix.h>
#include <LEDA/point.h>
#include <LEDA/polygon.h>
#include <LEDA/stream.h>
#include <LEDA/string.h>
#include <LEDA/vector.h>

#include <LEDA/REDEFINE_NAMES.h>

// The world places objects using configuration (model uses state).
// Each derived class can define Obst and Robot however it likes.
// Methods such as "GeomToPolygons" can give output in whatever 
//  format is requested.  So far, only list<polygon> is shown,
//  but we could imagine an OpenInventor model, Geomview model, etc.

//! Geometric models and collision detection methods.

/*!  These classes define the geometric representations of all obstacles in
the world, and of each part of the robot.  The methods allow planning
algorithms to determine whether any of the robot parts are in
collision with each other or with obstacles in the world.  Currently,
there are classes for 2D and 3D models, and two collision detection
methods: one uses LEDA primitives, and the other uses the PQP
collision detection package from the Univ. of North Carolina.  

A configuration vector specifies the positions and orientation of each
rigid body.
*/

class Geom {
 protected:
  string FilePath;
 public:
  //! Empty constructor in base class
  Geom(string path);

  //! Empty destructor
  virtual ~Geom() {};

  //! The number of rigid bodies in the geometry model
  int NumBodies;

  //! The dimension of the world geometry: 2 or 3
  int GeomDim;

  //! Return true if the robot(s) and obstacles are not in collision
  virtual bool CollisionFree(const vector &q) = 0; // Input is configuration

  //! Compute the distance of the closest point on the robot to the
  //! obstacle region.
  virtual double DistanceComp(const vector &q) = 0;  // Distance in world

  //! This is included for backwards compatibility, and will be removed one day
  virtual list<polygon> EnvironmentToLedaPolygons() = 0;

  //! This is included for backwards compatibility, and will be removed one day
  virtual list<polygon> RobotToLedaPolygons(const vector &q) = 0;

  //! Maximum displacement of geometry with respect to change in each variable
  vector MaxDeviates;

  //! Compute a vector based on q2-q1.  In R^n, the configurations are simply
  //! subtracted to make the vector.  This method exists to make things
  //! work correctly for other configuration-space topologies.
  virtual vector ConfigurationDifference(const vector &q1, 
					 const vector &q2); 
};


//! A point robot in a polygonal world.
class GeomLedaPolygonal: public Geom {
 protected:
  list<polygon> Obst;
 public:
  GeomLedaPolygonal(string path);
  virtual ~GeomLedaPolygonal() {};
  virtual bool CollisionFree(const vector &q); // Input is configuration
  virtual double DistanceComp(const vector &q);  // Distance in world
  // These two functions enable a "frame" to be displayed in using
  // LEDA polygons, even if Obst and Robot are represented using
  // other (including 3D) primitives).
  virtual list<polygon> EnvironmentToLedaPolygons();
  virtual list<polygon> RobotToLedaPolygons(const vector &q);
  // Other methods could be added here that are similar to the two
  // above, but output the world in different formats, such 
  // as: Geomview file, Open Inventor file, GL primitives, etc.
};


//! A rigid 2D robot in a polygonal world.
class GeomLedaPolygonalRigid: public GeomLedaPolygonal {
 private:
  list<polygon> Robot;
 public:
  GeomLedaPolygonalRigid(string path);
  virtual ~GeomLedaPolygonalRigid() {};
  virtual bool CollisionFree(const vector &q); // Input is configuration
  virtual double DistanceComp(const vector &q);  // Distance in world
  virtual list<polygon> RobotToLedaPolygons(const vector &q);
  virtual vector ConfigurationDifference(const vector &q1, 
					      const vector &q2); 
};


#include <LEDA/UNDEFINE_NAMES.h>

#endif
