//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#ifndef MSL_MODEL_H
#define MSL_MODEL_H

#include <LEDA/file.h>
#include <LEDA/list.h>
#include <LEDA/matrix.h>
#include <LEDA/point.h>
#include <LEDA/polygon.h>
#include <LEDA/random.h>
#include <LEDA/stream.h>
#include <LEDA/string.h>
#include <LEDA/vector.h>

#include <LEDA/REDEFINE_NAMES.h>

//! The incremental simulator model

/*!  The Model classes contain incremental simulators that model the
kinematics and dynamics of a variety of mechanical systems.  The
methods allow planning algorithms to compute the future system state,
given the current state, an interval of time, and a control input
applied over that interval.  (The planning algorithms select the
appropriate inputs using this information.)  Using object-oriented
class derivations, a wide variety of simulators can be included.  */

class Model {
 protected:
  //! The time interval to use for numerical integration (affects accuracy)
  double ModelDeltaT;

  //! The complete set of inputs
  list<vector> Inputs;

  //! Read the inputs from a file named Inputs
  void ReadInputs();

  //! Read the lowest possible state from a file named LowerState 
  void ReadLowerState();

  //! Read the upper state from a file named UpperState
  void ReadUpperState();

  //! Integrate xdot using 4th-order Runge-Kutta
  vector RungeKuttaIntegrate(const vector &x, const vector &u, const double &h);

  //! Integrate xdot using Euler integration
  vector EulerIntegrate(const vector &x, const vector &u, const double &h);
 public:

  //! This file path is used for all file reads
  string FilePath;

  //! Vector of minimum values for each state variable
  vector LowerState; 

  //! Vector of maximum values for each state variable
  vector UpperState;

  //! Vector of minimum values for each input variable
  vector LowerInput;

  //! Vector of maximum values for each input variable
  vector UpperInput;

  //! The dimension of the state space
  int StateDim;
  
  //! The dimension of the input space
  int InputDim;

  //! Empty constructor
  Model(string path);

  //! Empty destructor
  virtual ~Model() {};

  //! Return a list of inputs, which may depend on state
  virtual list<vector> GetInputs(const vector &x);

  //! The state transition equation, or equations of motion, xdot=f(x,u)
  virtual vector StateTransitionEquation(const vector &x, const vector &u) = 0;

  //! Test whether global state-space constraints are satisfied
  virtual bool Satisfied(const vector &x);

  //! Perform integration from state x, using input u, over time step h
  virtual vector Integrate(const vector &x, const vector &u, 
			   const double &h) = 0; 

  //! Linearly interpolate two state while respecting topology.
  /*! If a=0, then x1 is returned; if a=1, then x2 is returned.  All
      intermediate values of $a \in [0,1]$ yield intermediate states.
      This method is defined by Model.
  */
  virtual vector LinearInterpolate(const vector &x1, const vector &x2, 
				   const double &a);  // Depends on topology

  //! Compute a vector based on x2-x1.  In R^n, the states are simply
  //! subtracted to make the vector.  This method exists to make things
  //! work correctly for other state-space topologies.
  virtual vector StateDifference(const vector &x1, const vector &x2); 

  // Conversions
  //! A method that converts a Model state in to a Geom configuration
  virtual vector StateToConfiguration(const vector &x);

  //! A method that converts a state into a 2D LEDA point
  virtual point  StateToLedaPoint(const vector &x);

  //! A distance metric, which is Euclidean in the base class
  virtual double Metric(const vector &x1, const vector &x2); 

  // The following are used by optimization methods.  They are empty by
  // default because regular planners don't need them.  These could later
  // go in a derived class for optimization problems, but are left here
  // so that "regular" models can be converted to optimization models
  // by overriding these methods.

  //! Partial with respect to x of the state transition equation
  virtual void Partialf_x(const vector &x, const vector &u, matrix & m) {};

  //! Partial with respect to u of the state transition equation
  virtual void Partialf_u(const vector &x, const vector &u, matrix & m) {};

  //! A cost or loss function, to be used in optimization problems
  virtual void L(const vector &x, const vector &u, double &l) {};

  //! Partial of the loss with respect to x
  virtual void PartialL_x(const vector &x, const vector &u, matrix & m) {};

  //! Partial of the loss with respect to u
  virtual void PartialL_u(const vector &x, const vector &u, matrix & m) {};

  //! The final-state cost or loss
  virtual void Phi(const vector &x, const vector &u, 
		   const vector &goalstate, double &phi) {};

  //! Partial of the final-state loss with respect to x
  virtual void PartialPhi_x(const vector &x, const vector &u, 
			    const vector &goalstate, 
			    matrix & m) {};

  //! Partial of the final-state loss with respect to u
  virtual void PartialPhi_t(const vector &x, const vector &u, 
			    const vector &goalstate, 
			    matrix & m) {};

  //! An error vector that compares a goal state to a given state
  virtual void Psi(const vector &x, const vector &goalstate, vector& psi) {};

  //! Partial of the error vector with respect to x
  virtual void PartialPsi_x(const vector &x, const vector &u, matrix & m) {};

  //! Partial of the error vector with respect to time
  virtual void PartialPsi_t(const vector &x, const vector &u, matrix & m) {};

};

#include <LEDA/UNDEFINE_NAMES.h>

#endif
