//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#include <fstream.h>
#include <math.h>

#include "model2d.h"
#include "defs.h"

#include <LEDA/REDEFINE_NAMES.h>

#define NUM_INPUTS_2DPOINT 50
#define NUM_INPUTS_2DRIGID 8


// *********************************************************************
// *********************************************************************
// CLASS:     Model2D
// 
// *********************************************************************
// *********************************************************************



Model2D::Model2D(string path = ""):Model(path) {
}


vector Model2D::StateToConfiguration(const vector &x) 
{
  vector q(3);

  q[0] = x[0]; q[1] = x[1]; q[2] = 0.0;

  return q;
}



// *********************************************************************
// *********************************************************************
// CLASS:     Model2DPoint
// 
// *********************************************************************
// *********************************************************************


// Constructor
Model2DPoint::Model2DPoint(string path = ""):Model2D(path) {
  double theta;

  StateDim = 2;
  InputDim = 2;

  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState = vector(0.0,0.0);
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState = vector(100.0,100.0);
  }

  // Make the list of Inputs
  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    for (theta = 0.0; theta < 2.0*PI; theta += 2.0*PI/NUM_INPUTS_2DPOINT) {
      Inputs.push_back(vector(cos(theta),sin(theta)));
    }
  }
}




vector Model2DPoint::StateTransitionEquation(const vector &x, const vector &u) {

  vector dx(2);

  dx = u;
  return dx;
}



double Model2DPoint::Metric(const vector &x1, const vector &x2) {

  double rho;

  rho = (x1 - x2).length();

  return rho;
}





vector Model2DPoint::Integrate(const vector &x, const vector &u, const double &h)
{
  return EulerIntegrate(x,u,h);
}



// *********************************************************************
// *********************************************************************
// CLASS:     Model::Model2DPointCar
// 
// *********************************************************************
// *********************************************************************


// Constructor
Model2DPointCar::Model2DPointCar(string path = ""):Model2DPoint(path) {
  double alpha;

  StateDim = 3;
  InputDim = 2;

  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState = vector(0.0,0.0,0.0);
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState = vector(100.0,100.0,2.0*PI);
  }

  MaxSteeringAngle = PI/6.0;
  CarLength = 4.0;

  // Make the list of Inputs
  Inputs.clear();  // Otherwise its parent constructor will make some inputs
  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    for (alpha = -MaxSteeringAngle; alpha <= MaxSteeringAngle; 
	 alpha += 2.0*MaxSteeringAngle/6.0) {
      Inputs.push_back(vector(1.0,alpha));
      Inputs.push_back(vector(-1.0,alpha)); 
    }
  }
}



vector Model2DPointCar::StateTransitionEquation(const vector &x, const vector &u) {

  vector dx(3);

  dx[0] = u[0]*cos(x[2]);
  dx[1] = u[0]*sin(x[2]);
  dx[2] = u[0]*tan(u[1])/CarLength;
  return dx;
}



double Model2DPointCar::Metric(const vector &x1, const vector &x2) {

  double rho,dtheta;
  
  dtheta = min(fabs(x1[2]-x2[2]),2.0*PI - fabs(x1[2]-x2[2]));

  rho = sqrt(sqr(x1[0] - x2[0]) + sqr(x1[1] - x2[1]) + 50.0/PI*sqr(dtheta));

  return rho;
}



vector Model2DPointCar::Integrate(const vector &x, const vector &u, 
				  const double &h) {
  vector nx(3);

  nx = RungeKuttaIntegrate(x,u,h);

  // Make sure the S^1 topology is preserved for 2D rotation
  if (nx[2] > 2.0*PI)
    nx[2] -= 2.0*PI;
  if (nx[2] < 0.0)
    nx[2] += 2.0*PI;

  return nx;
}



// *********************************************************************
// *********************************************************************
// CLASS:     Model2DRigid
// 
// *********************************************************************
// *********************************************************************



// Constructor
Model2DRigid::Model2DRigid(string path = ""):Model2D(path) {
  double theta;

  StateDim = 3;
  InputDim = 3;

  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState = vector(0.0,0.0,0.0);
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState = vector(100.0,100.0,2.0*PI);
  }

  // Make the list of Inputs
  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    for (theta = 0.0; theta < 2.0*PI; theta += 2.0*PI/NUM_INPUTS_2DRIGID) {
      Inputs.push_back(vector(cos(theta),sin(theta),0.0));
      //Inputs.push_back(vector(cos(theta),sin(theta),-0.1));
      //Inputs.push_back(vector(cos(theta),sin(theta),0.1));
    }
    Inputs.push_back(vector(0.0,0.0,-0.1));
    Inputs.push_back(vector(0.0,0.0,0.1));
  }

}



vector Model2DRigid::StateTransitionEquation(const vector &x, const vector &u) {

  vector dx(3);

  dx = u;
  return dx;
}



double Model2DRigid::Metric(const vector &x1, const vector &x2) {

  double fd = fabs(x1[2]-x2[2]);
  double dtheta = min(fd,2.0*PI - fd);

  return sqrt(sqr(x1[0] - x2[0]) + sqr(x1[1] - x2[1]) + sqr(50.0/PI*dtheta));

}



// Handle S^1 topology properly (for rotation)
vector Model2DRigid::LinearInterpolate(const vector &x1, const vector &x2, 
				       const double &a) {

  vector v;

  v = (1.0-a)*x1 + a*x2;

  if (fabs(x2[2] - x1[2]) > PI) {
    if (x1[2] > x2[2])
      v[2] = (1.0-a)*x1[2] + a*(x2[2]+2.0*PI);
    else
      v[2] = (1.0-a)*(x1[2]+2.0*PI) + a*x2[2];
  }

  if (v[2] > 2.0*PI)
    v[2] -= 2.0*PI;

  return v;
}



vector Model2DRigid::Integrate(const vector &x, const vector &u, const double &h)
{
  vector nx(3);

  nx = RungeKuttaIntegrate(x,u,h);

  // Make sure the S^1 topology is preserved for 2D rotation
  if (nx[2] > 2.0*PI)
    nx[2] -= 2.0*PI;
  if (nx[2] < 0.0)
    nx[2] += 2.0*PI;

  return nx;
}


vector Model2DRigid::StateToConfiguration(const vector &x) 
{
  vector q(3);

  q[0] = x[0]; q[1] = x[1]; q[2] = x[2];

  return q;
}


// *********************************************************************
// *********************************************************************
// CLASS:     Model2DRigidCar
// 
// *********************************************************************
// *********************************************************************


// Constructor
Model2DRigidCar::Model2DRigidCar(string path = ""):Model2DRigid(path) {
  double alpha;

  StateDim = 3;
  InputDim = 2;

  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState = vector(0.0,0.0,0.0);
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState = vector(100.0,100.0,2.0*PI);
  }

  MaxSteeringAngle = PI/12.0;
  CarLength = 2.0;

  // Make the list of Inputs
  Inputs.clear();  // Otherwise its parent constructor will make some inputs
  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    for (alpha = -MaxSteeringAngle; alpha <= MaxSteeringAngle; 
	 alpha += 2.0*MaxSteeringAngle/6.0) {
      Inputs.push_back(vector(1.0,alpha)); 
      Inputs.push_back(vector(-1.0,alpha)); 
    }
  }
}


vector Model2DRigidCar::StateTransitionEquation(const vector &x, const vector &u) {

  vector dx(3);
  dx[0] = u[0]*cos(x[2]);
  dx[1] = u[0]*sin(x[2]);
  dx[2] = u[0]*tan(u[1])/CarLength;
  return dx;
}


// *********************************************************************
// *********************************************************************
// CLASS:     Model2DRigidCarForward
// 
// *********************************************************************
// *********************************************************************

Model2DRigidCarForward::Model2DRigidCarForward(string path = ""):Model2DRigidCar(path) {
  double alpha;

  StateDim = 3;
  InputDim = 2;

 // Defaults are inherited
  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  if (is_file(FilePath+"UpperState"))
    ReadUpperState();

  // Make the list of Inputs
  Inputs.clear();  // Otherwise its parent constructor will make some inputs
  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    for (alpha = -MaxSteeringAngle; alpha <= MaxSteeringAngle; 
	 alpha += 2.0*MaxSteeringAngle/6.0) {
      Inputs.push_back(vector(1.0,alpha)); 
    }
  }
}


// *********************************************************************
// *********************************************************************
// CLASS:     Model2DRigidCarSmooth
//  Smooth steering 
// *********************************************************************
// *********************************************************************


Model2DRigidCarSmooth::Model2DRigidCarSmooth(string path = ""):Model2DRigidCar(path) {

  StateDim = 4;
  InputDim = 2;
  SteeringSpeed = 0.05;  // Default 0.05
  
  LowerState = vector(4);
  UpperState = vector(4);

  if (is_file(FilePath+"SteeringSpeed")) {
    file_istream fin(FilePath + "SteeringSpeed");
    fin >> SteeringSpeed;
    fin.close();
  }
  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState[0] = 0.0; LowerState[1] = 0.0; LowerState[2] = 0.0; 
    LowerState[3] = -MaxSteeringAngle;
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState[0] = 100.0; UpperState[1] = 100.0; UpperState[2] = 2.0*PI; 
    UpperState[3] = MaxSteeringAngle;
  }

  // Make the list of Inputs
  Inputs.clear();  // Otherwise its parent constructor will make some inputs
  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    Inputs.push_back(vector(1.0,0.0)); // Keep the steering angle fixed
    Inputs.push_back(vector(-1.0,0.0)); // Keep the steering angle fixed
    Inputs.push_back(vector(1.0,SteeringSpeed)); 
    Inputs.push_back(vector(-1.0,SteeringSpeed)); 
    Inputs.push_back(vector(1.0,-SteeringSpeed)); 
    Inputs.push_back(vector(-1.0,-SteeringSpeed)); 
   }
}



vector Model2DRigidCarSmooth::StateTransitionEquation(const vector &x, const vector &u) {

  vector dx(4);

  dx[0] = u[0]*cos(x[2]);
  dx[1] = u[0]*sin(x[2]);
  dx[2] = u[0]*tan(x[3])/CarLength;
  dx[3] = u[1];

  //cout << "DX: " << dx << "\n";

  return dx;
}




vector Model2DRigidCarSmooth::Integrate(const vector &x, const vector &u, const double &h)
{
  return RungeKuttaIntegrate(x,u,h);
}


double Model2DRigidCarSmooth::Metric(const vector &x1, const vector &x2) {

  double rho,dphi,dtheta;

  dphi = min(fabs(x1[3]-x2[3]),2.0*PI - fabs(x1[3]-x2[3]));
  dtheta = min(fabs(x1[2]-x2[2]),2.0*PI - fabs(x1[2]-x2[2]));

  rho = sqrt(sqr(x1[0] - x2[0]) + sqr(x1[1] - x2[1]) +
    sqr(2.0/PI*dphi) +
    sqr(50.0/PI*dtheta));

  return rho;
}


vector Model2DRigidCarSmooth::StateToConfiguration(const vector &x) 
{
  vector q(3);

  q[0] = x[0]; q[1] = x[1]; q[2] = x[2];

  return q;
}


bool Model2DRigidCarSmooth::Satisfied(const vector &x) 
{
  return ((x[3] < UpperState[3])&& // Steering is too sharp!
	  (x[3] > LowerState[3]));
}


// *********************************************************************
// *********************************************************************
// CLASS:     Model2DRigidCarSmoothTrailer
//  Smooth steering 
// *********************************************************************
// *********************************************************************


Model2DRigidCarSmoothTrailer::Model2DRigidCarSmoothTrailer(string path = ""):Model2DRigidCarSmooth(path) {
  
  StateDim = 5;
  InputDim = 2;
  HitchLength = 10.0;
  HitchMaxAngle = PI/2.0;  // From 0 to PI (PI allows 360 spin of trailer)

  LowerState = vector(5);
  UpperState = vector(5);

  if (is_file(FilePath+"HitchLength")) {
    file_istream fin(FilePath + "HitchLength");
    fin >> HitchLength;
    fin.close();
  }
  if (is_file(FilePath+"HitchMaxAngle")) {
    file_istream fin(FilePath + "HitchMaxAngle");
    fin >> HitchMaxAngle;
    fin.close();
  }

  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState[0] = 0.0; LowerState[1] = 0.0; LowerState[2] = 0.0; 
    LowerState[3] = -MaxSteeringAngle; LowerState[4] = 0.0;
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState[0] = 100.0; UpperState[1] = 100.0; UpperState[2] = 2.0*PI; 
    UpperState[3] = MaxSteeringAngle; UpperState[4] = 2.0*PI;
  }

  if (is_file(FilePath+"Inputs"))
    ReadInputs();

}



vector Model2DRigidCarSmoothTrailer::StateTransitionEquation(const vector &x, const vector &u) {

  vector dx(5);

  dx[0] = u[0]*cos(x[2]);
  dx[1] = u[0]*sin(x[2]);
  dx[2] = u[0]*tan(x[3])/CarLength;
  dx[3] = u[1];
  dx[4] = u[0]*sin(x[2] - x[4])/HitchLength;
  return dx;
}




double Model2DRigidCarSmoothTrailer::Metric(const vector &x1, const vector &x2) {

  double rho,dphi,dtheta,dtheta1;

  dphi = min(fabs(x1[3]-x2[3]),2.0*PI - fabs(x1[3]-x2[3]));
  dtheta = min(fabs(x1[2]-x2[2]),2.0*PI - fabs(x1[2]-x2[2]));
  dtheta1 = min(fabs(x1[4]-x2[4]),2.0*PI - fabs(x1[4]-x2[4]));

  rho = sqrt(sqr(x1[0] - x2[0]) + 
	     sqr(x1[1] - x2[1]) +
	     sqr(2.0/PI*dphi) +
	     sqr(5.0/PI*dtheta) +
	     sqr(5.0/PI*dtheta1));

  return rho;
}


vector Model2DRigidCarSmoothTrailer::StateToConfiguration(const vector &x) 
{
  vector q(6);  // Two bodies

  // The car config
  q[0] = x[0]; q[1] = x[1]; q[2] = x[2];

  // The trailer config
  q[3] = -cos(x[4])*HitchLength+x[0];
  q[4] = -sin(x[4])*HitchLength+x[1];
  q[5] = x[4];

  return q;
}


bool Model2DRigidCarSmoothTrailer::Satisfied(const vector &x) 
{
  return ((x[3] < UpperState[3])&& // Steering is too sharp!
	  (x[3] > LowerState[3])&&
	  (cos(x[2]-x[4]) >= cos(HitchMaxAngle)));
}


// *********************************************************************
// *********************************************************************
// CLASS:     Model2DRigidCarSmooth2Trailers
//  Smooth steering 
// *********************************************************************
// *********************************************************************


Model2DRigidCarSmooth2Trailers::Model2DRigidCarSmooth2Trailers(string path = ""):Model2DRigidCarSmoothTrailer(path) {
  
  StateDim = 6;
  InputDim = 2;
  Hitch2Length = 10.0;
  Hitch2MaxAngle = PI/2.0;  // From 0 to PI (PI allows 360 spin of trailer)

  LowerState = vector(6);
  UpperState = vector(6);

  if (is_file(FilePath+"Hitch2Length")) {
    file_istream fin(FilePath + "Hitch2Length");
    fin >> Hitch2Length;
    fin.close();
  }
  if (is_file(FilePath+"Hitch2MaxAngle")) {
    file_istream fin(FilePath + "Hitch2MaxAngle");
    fin >> Hitch2MaxAngle;
    fin.close();
  }
  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState[0] = 0.0; LowerState[1] = 0.0; LowerState[2] = 0.0; 
    LowerState[3] = -MaxSteeringAngle; LowerState[4] = 0.0;
    LowerState[5] = 0.0;
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState[0] = 100.0; UpperState[1] = 100.0; UpperState[2] = 2.0*PI; 
    UpperState[3] = MaxSteeringAngle; UpperState[4] = 2.0*PI;
    UpperState[5] = 2.0*PI;
  }

  if (is_file(FilePath+"Inputs"))
    ReadInputs();

}



vector Model2DRigidCarSmooth2Trailers::StateTransitionEquation(const vector &x, const vector &u) {

  vector dx(6);

  dx[0] = u[0]*cos(x[2]);
  dx[1] = u[0]*sin(x[2]);
  dx[2] = u[0]*tan(x[3])/CarLength;
  dx[3] = u[1];
  dx[4] = u[0]*sin(x[2] - x[4])/HitchLength;
  dx[5] = u[0]*cos(x[2] - x[4])*sin(x[4] - x[5])/Hitch2Length;
  return dx;
}



double Model2DRigidCarSmooth2Trailers::Metric(const vector &x1, const vector &x2) {

  double rho,dphi,dtheta,dtheta1,dtheta2;

  dphi = min(fabs(x1[3]-x2[3]),2.0*PI - fabs(x1[3]-x2[3]));
  dtheta = min(fabs(x1[2]-x2[2]),2.0*PI - fabs(x1[2]-x2[2]));
  dtheta1 = min(fabs(x1[4]-x2[4]),2.0*PI - fabs(x1[4]-x2[4]));
  dtheta2 = min(fabs(x1[5]-x2[5]),2.0*PI - fabs(x1[5]-x2[5]));

  rho = sqrt(sqr(x1[0] - x2[0]) + sqr(x1[1] - x2[1]) +
    sqr(2.0/PI*dphi) +
    sqr(5.0/PI*dtheta) +
    sqr(5.0/PI*dtheta1) +
    sqr(5.0/PI*dtheta2));

  return rho;
}


vector Model2DRigidCarSmooth2Trailers::StateToConfiguration(const vector &x) 
{
  vector q(9);  // Three bodies

  // The car config
  q[0] = x[0]; q[1] = x[1]; q[2] = x[2];

  // The 1st trailer config
  q[3] = -cos(x[4])*HitchLength+x[0];
  q[4] = -sin(x[4])*HitchLength+x[1];
  q[5] = x[4];

  // The 2nd trailer config
  q[6] = -cos(x[5])*Hitch2Length+q[3];
  q[7] = -sin(x[5])*Hitch2Length+q[4];
  q[8] = x[5];

  return q;
}


bool Model2DRigidCarSmooth2Trailers::Satisfied(const vector &x) 
{
  return ((x[3] < UpperState[3])&& // Steering is too sharp!
	  (x[3] > LowerState[3])&&
	  (cos(x[2]-x[4]) >= cos(HitchMaxAngle))&&
	  (cos(x[4]-x[5]) >= cos(Hitch2MaxAngle)));
}


// *********************************************************************
// *********************************************************************
// CLASS:     Model2DRigidCarSmooth3Trailers
//  Smooth steering 
// *********************************************************************
// *********************************************************************


Model2DRigidCarSmooth3Trailers::Model2DRigidCarSmooth3Trailers(string path = ""):Model2DRigidCarSmooth2Trailers(path) {
  
  StateDim = 7;
  InputDim = 2;
  Hitch3Length = 10.0;
  Hitch3MaxAngle = PI/2.0;  // From 0 to PI (PI allows 360 spin of trailer)

  LowerState = vector(7);
  UpperState = vector(7);

  if (is_file(FilePath+"Hitch3Length")) {
    file_istream fin(FilePath + "Hitch3Length");
    fin >> Hitch3Length;
    fin.close();
  }
  if (is_file(FilePath+"Hitch3MaxAngle")) {
    file_istream fin(FilePath + "Hitch3MaxAngle");
    fin >> Hitch3MaxAngle;
    fin.close();
  }
  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState[0] = 0.0; LowerState[1] = 0.0; LowerState[2] = 0.0; 
    LowerState[3] = -MaxSteeringAngle; LowerState[4] = 0.0;
    LowerState[5] = 0.0; LowerState[6] = 0.0;
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState[0] = 100.0; UpperState[1] = 100.0; UpperState[2] = 2.0*PI; 
    UpperState[3] = MaxSteeringAngle; UpperState[4] = 2.0*PI;
    UpperState[5] = 2.0*PI; UpperState[6] = 2.0*PI;
  }

  if (is_file(FilePath+"Inputs"))
    ReadInputs();

}



vector Model2DRigidCarSmooth3Trailers::StateTransitionEquation(const vector &x, const vector &u) {

  vector dx(7);

  dx[0] = u[0]*cos(x[2]);
  dx[1] = u[0]*sin(x[2]);
  dx[2] = u[0]*tan(x[3])/CarLength;
  dx[3] = u[1];
  dx[4] = u[0]*sin(x[2] - x[4])/HitchLength;
  dx[5] = u[0]*cos(x[2] - x[4])*sin(x[4] - x[5])/Hitch2Length;
  dx[6] = u[0]*cos(x[2] - x[4])*cos(x[4] - x[5])*sin(x[5]-x[6])/Hitch3Length;
  return dx;
}




double Model2DRigidCarSmooth3Trailers::Metric(const vector &x1, const vector &x2) {

  double rho,dphi,dtheta,dtheta1,dtheta2,dtheta3;

  dphi = min(fabs(x1[3]-x2[3]),2.0*PI - fabs(x1[3]-x2[3]));
  dtheta = min(fabs(x1[2]-x2[2]),2.0*PI - fabs(x1[2]-x2[2]));
  dtheta1 = min(fabs(x1[4]-x2[4]),2.0*PI - fabs(x1[4]-x2[4]));
  dtheta2 = min(fabs(x1[5]-x2[5]),2.0*PI - fabs(x1[5]-x2[5]));
  dtheta3 = min(fabs(x1[6]-x2[6]),2.0*PI - fabs(x1[6]-x2[6]));

  rho = sqrt(sqr(x1[0] - x2[0]) + sqr(x1[1] - x2[1]) +
    sqr(2.0/PI*dphi) +
    sqr(5.0/PI*dtheta) +
    sqr(5.0/PI*dtheta1) +
    sqr(5.0/PI*dtheta2) +
    sqr(5.0/PI*dtheta3));

  return rho;
}



vector Model2DRigidCarSmooth3Trailers::StateToConfiguration(const vector &x) 
{
  vector q(12);  // Four bodies

  // The car config
  q[0] = x[0]; q[1] = x[1]; q[2] = x[2];

  // The 1st trailer config
  q[3] = -cos(x[4])*HitchLength+x[0];
  q[4] = -sin(x[4])*HitchLength+x[1];
  q[5] = x[4];

  // The 2nd trailer config
  q[6] = -cos(x[5])*Hitch2Length+q[3];
  q[7] = -sin(x[5])*Hitch2Length+q[4];
  q[8] = x[5];

  // The 3rd trailer config
  q[9]  = -cos(x[6])*Hitch3Length+q[6];
  q[10] = -sin(x[6])*Hitch3Length+q[7];
  q[11] = x[6];

  return q;
}


bool Model2DRigidCarSmooth3Trailers::Satisfied(const vector &x) 
{
  return ((x[3] < UpperState[3])&& // Steering is too sharp!
	  (x[3] > LowerState[3])&&
	  (cos(x[2]-x[4]) >= cos(HitchMaxAngle))&&
	  (cos(x[4]-x[5]) >= cos(Hitch2MaxAngle))&&
	  (cos(x[5]-x[6]) >= cos(Hitch3MaxAngle)));
}


// *********************************************************************
// *********************************************************************
// CLASS:     Model2DRigidDyncar
//  
// *********************************************************************
// *********************************************************************

// Constructor
Model2DRigidDyncar::Model2DRigidDyncar(string path = ""):Model2DRigid(path) {
  double alpha;
  vector v;

  StateDim = 5;
  InputDim = 1;
  Mass = 100.0;
  CAF = 17000.0;
  CAR = 20000.0;
  Adist = 4.0;
  Bdist = 5.0;
  Izz = 1600.0;

  WorldScale = 0.1;
  Speed = 88.0;   // Feet per sec
  
  LowerState = vector(5);
  UpperState = vector(5);

  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState[0] = -50.0; LowerState[1] = -5.0; LowerState[2] = 0.0; 
    LowerState[3] = -1000.0; LowerState[4] = 0.0;
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState[0] = 50.0; UpperState[1] = 5.0; UpperState[2] = 1000.0; 
    UpperState[3] = 0.0; UpperState[4] = 2.0*PI;
  }

  MaxSteeringAngle = 0.6;

  // Make the list of 1D Inputs
  Inputs.clear();  // Otherwise its parent constructor will make some inputs
  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    for (alpha = -MaxSteeringAngle; alpha <= MaxSteeringAngle; 
	 alpha += 2.0*MaxSteeringAngle/24.0) {
      v = vector(1); v[0] = alpha;
      Inputs.push_back(v);
    }
  }
}


// This is xdot = f(x,u) for a 5dof state-space, and 1dof input
//  Model taken from Jim Bernard...
// alphaf = (v + a*r)/u - deltaf
// alphar = (v-b*r)/u
// FYF = -CAF*alphaf
// FYR = -CAR*alphar
// vdot = -u*r  + (FYF + FYR)/m
// rdot = (FYF*a - FYR*b)/Izz
// Xdot = u*cos(psi) - v*sin(psi)
// Ydot = u*sin(spi) + v*cos(psi)
// psidot = r
// 
// m = mass of car, say about 100 slugs
// CAF = front cornering stiffness in pounds per radiaof the tires, say about
// 17000
// CAR = rear cornering stiffness, say about 20000
// a is dist from mass center to front tires, say 4 feet
// b is dist from mass center to rear tires, say 5 feet
// Izz is yaw moment of intertia, say about 1600 slug ft**2
// u is forward speed which is assumed constant, input is in feet/sec
// delta is your input, it is the steer angle of the tires in radians.

vector Model2DRigidDyncar::StateTransitionEquation(const vector &x, const vector &u)
{
  double alphaf,alphar,fyf,fyr,v,r,psi;
  vector dx(5);

  v = x[0]; r = x[1]; psi = x[4];

  alphaf = (v + Adist * r) / Speed - u[0];
  alphar = (v - Bdist * r) / Speed;
  fyf = -CAF * alphaf;
  fyr = -CAR * alphar;

  /* Transfer the velocity */
  dx[0] = -Speed * r  + (fyf + fyr) / Mass;
  dx[1] = (fyf * Adist - fyr * Bdist) / Izz;
  dx[2] = Speed * cos(psi) - v * sin(psi);
  dx[3] = Speed * sin(psi) + v * cos(psi);
  dx[4] = r;

  //cout << "x: " << x << "  Dx: " << dx << "  u: " << u[0] << "\n";

  return dx;
}



double Model2DRigidDyncar::Metric(const vector &x1, const vector &x2) {
  double d;

  // Position difference
  d =  sqr((x1[2] - x2[2]) / (UpperState[2] - LowerState[2]));
  d += sqr((x1[3] - x2[3]) / (UpperState[3] - LowerState[3]));

  // Orientation difference
  d += sqr(min(fabs(x1[4]-x2[4]),2.0*PI - fabs(x1[4]-x2[4]))/2.0/PI);

  // Velocities
  d += sqr((x1[0] - x2[0]) / (UpperState[0] - LowerState[0]));
  d += sqr((x1[1] - x2[1]) / (UpperState[1] - LowerState[1]));


  return sqrt(d);
}



vector Model2DRigidDyncar::Integrate(const vector &x, const vector &u, const double &h)
{
  return RungeKuttaIntegrate(x,u,h);
}



point Model2DRigidDyncar::StateToLedaPoint(const vector &x) {
  return point(x[2]*WorldScale,-x[3]*WorldScale);
}



// NOTE: This neglects the S^1 effect of orientation!!
vector Model2DRigidDyncar::StateToConfiguration(const vector &x) {
  return vector(x[2]*WorldScale,-x[3]*WorldScale,2.0*PI-x[4]); 
}


vector Model2DRigidDyncar::LinearInterpolate(const vector &x1, 
					     const vector &x2, 
					     const double &a) {

  vector v;

  v = (1.0-a)*x1 + a*x2;

  if (fabs(x2[4] - x1[4]) > PI) {
    if (x1[4] > x2[4])
      v[4] = (1.0-a)*x1[4] + a*(x2[4]+2.0*PI);
    else
      v[4] = (1.0-a)*(x1[4]+2.0*PI) + a*x2[4];
  }

  if (v[4] > 2.0*PI)
    v[4] -= 2.0*PI;

  return v;
}



// *********************************************************************
// *********************************************************************
// CLASS:     Model2DRigidDyncarNtire
//  
// *********************************************************************
// *********************************************************************

// Constructor
Model2DRigidDyncarNtire::Model2DRigidDyncarNtire(string path = ""):Model2DRigidDyncar(path) {
  double alpha;
  vector v;

  StateDim = 5;
  InputDim = 2;

  // These are the exact parameters from Ric's model
  Mass = 3518.0/32.2;
  Adist = 0.45*100.5/12.0;
  Bdist = 0.55*100.5/12.0;
  Izz = 0.25*Mass*(Adist+Bdist)*(Adist+Bdist);

  // Constants for the nonlinear tire model
  Mu = 0.85;
  Nf = Mass*32.2*0.55;
  Nr = Mass*32.2*0.45;

  // The defaults are inherited from parent class
  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  if (is_file(FilePath+"UpperState"))
    ReadUpperState();


  // Make the list of 2D Inputs
  // Apparently, this models allows rear-wheel steering
  //    This option is set to zero for now...hence v[1]=0.0
  Inputs.clear();  // Otherwise its parent constructor will make some inputs
  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    for (alpha = -MaxSteeringAngle; alpha <= MaxSteeringAngle; 
	 alpha += 2.0*MaxSteeringAngle/24.0) {
      v = vector(2); v[0] = alpha; v[1] = 0.0;
      Inputs.push_back(v);
    }
  }
}


// The is the model from Jim Bernard and Ric Menendez
vector Model2DRigidDyncarNtire::StateTransitionEquation(const vector &x, const vector &u)
{
  double alphaf,alphar,fyf,fyr,v,r,psi;
  double talff,talfr,xiblf,xiblr;
  vector dx(5);
  

  v = x[0]; r = x[1]; psi = x[4];

  alphaf = (v + Adist * r) / Speed - u[0];
  alphar = (v - Bdist * r) / Speed - u[1];

  // Here is where the nonlinear tire model is used
  //  The result is new values for fyf and fyr
  talff = tan(fabs(alphaf));
  talfr = tan(fabs(alphar));
  xiblf = (CAF*talff == 0) ? 
    INFINITY :
    Mu*Nf/(2.0*CAF*talff);
  xiblr = (CAR*talfr == 0) ? 
    INFINITY :
    Mu*Nr/(2.0*CAR*talfr);
  fyf = (xiblf >= 1.0) ?
    CAF*talff :
    Mu*Nf*(1.0-0.5*xiblf);
  fyr = (xiblr >= 1.0) ?
    CAR*talfr :
    Mu*Nr*(1.0-0.5*xiblr);
  fyf = (alphaf > 0) ? -1.0*fabs(fyf) : fabs(fyf);
  fyr = (alphar > 0) ? -1.0*fabs(fyr) : fabs(fyr);

  //cout << "talff: " << talff << "  xiblf: " << xiblf << " fyf: " << fyf << "\n";

  /* Transfer the velocity */
  dx[0] = -Speed * r  + (fyf + fyr) / Mass;
  dx[1] = (fyf * Adist - fyr * Bdist) / Izz;
  dx[2] = Speed * cos(psi) - v * sin(psi);
  dx[3] = Speed * sin(psi) + v * cos(psi);
  dx[4] = r;

  //cout << "x: " << x << "  Dx: " << dx << "  u: " << u[0] << "\n";

  return dx;
}







// *********************************************************************
// *********************************************************************
// CLASS:     Model2DRigidLander
//  
// *********************************************************************
// *********************************************************************

// Constructor
Model2DRigidLander::Model2DRigidLander(string path = ""):Model2DRigid(path) {
  vector v;

  StateDim = 4;
  InputDim = 1;
  Mass = 1.0;
  G = 1.568;  // Accel of gravity on moon (use 9.8 for earth)
  Fs = 10.0;
  Fu = 20.0;

  LowerState = vector(4);
  UpperState = vector(4);

  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState[0] = 0.0; LowerState[1] = 0.0; LowerState[2] = -10.0; 
    LowerState[3] = -10.0; 
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState[0] = 100.0; UpperState[1] = 100.0; UpperState[2] = 10.0; 
    UpperState[3] = 10.0; 
  }

  // Make the list of 1D Inputs
  Inputs.clear();  // Otherwise its parent constructor will make some inputs
  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    v = vector(1); v[0] = 0;
    Inputs.push_back(v);
    v = vector(1); v[0] = 1;
    Inputs.push_back(v);
    v = vector(1); v[0] = 2;
    Inputs.push_back(v);
    v = vector(1); v[0] = 3;
    Inputs.push_back(v);
  }
}



vector Model2DRigidLander::Integrate(const vector &x, const vector &u, const double &h)
{
  return RungeKuttaIntegrate(x,u,h);
}



vector Model2DRigidLander::StateToConfiguration(const vector &x) {
  return vector(x[0],x[1],0.0); // Yield a zero rotation every time
}


vector Model2DRigidLander::StateTransitionEquation(const vector &x, const vector &u)
{
  vector dx(4);

  /* Transfer the velocity */
  dx[0] = x[2];
  dx[1] = x[3];
  dx[2] = 0.0;
  if (u[0] == 1) 
    dx[2] = Fs;
  if (u[0] == 3) 
    dx[2] = -Fs;
  dx[3] = -Mass*G;
  if (u[0] == 2) 
    dx[3] += Fu;

  //cout << "x: " << x << "  Dx: " << dx << "  u: " << u[0] << "\n";

  return dx;
}





double Model2DRigidLander::Metric(const vector &x1, const vector &x2) {
  double d = 0.0;
  int i;

  for (i = 0; i < 4; i++) {
    d += sqrt(sqr(x1[i] - x2[i]) / (UpperState[i] - LowerState[i]));
  }
  
  //cout << "x1: " << x1 << "  x2: " << x2 << "   Metric: " << d << "\n";

  return d;
}




// *********************************************************************
// *********************************************************************
// CLASS:     Model2DRigidMulti
// 
// *********************************************************************
// *********************************************************************


// Constructor
Model2DRigidMulti::Model2DRigidMulti(string path = ""):Model2DRigid(path) {
  vector u;
  int i,j;

  file_istream fin(FilePath + "NumBodies");
  fin >> NumBodies;

  StateDim = 3*NumBodies;
  InputDim = 3*NumBodies;

  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState = vector(StateDim);
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState = vector(StateDim);
  }

  u = vector(StateDim);
  Inputs.clear();
  for (i = 0; i < StateDim; i++) {
    for (j = 0; j < StateDim; j++)
      u[j] = (i==j) ? 1.0 : 0.0; 
    Inputs.push(u);
    for (j = 0; j < StateDim; j++)
      u[j] = (i==j) ? -1.0 : 0.0; 
    Inputs.push(u);
  }
}




double Model2DRigidMulti::Metric(const vector &x1, const vector &x2) {

  double d,fd,dtheta;
  int i;

  d = 0.0;

  for (i = 0; i < NumBodies; i++) {
    fd = fabs(x1[3*i+2]-x2[3*i+2]);
    dtheta = min(fd,2.0*PI - fd);
    d += sqr(x1[3*i] - x2[3*i]);
    d += sqr(x1[3*i+1] - x2[3*i+1]);
    d += sqr(dtheta);
  }

  return sqrt(d);
}



vector Model2DRigidMulti::LinearInterpolate(const vector &x1, const vector &x2, const double &a){
  vector v;
  int i;

  v = (1.0-a)*x1 + a*x2;

  for (i = 0; i < NumBodies; i++) {

    if (fabs(x2[3*i+2] - x1[3*i+2]) > PI) {
      if (x1[3*i+2] > x2[3*i+2])
	v[3*i+2] = (1.0-a)*x1[3*i+2] + a*(x2[3*i+2]+2.0*PI);
      else
	v[3*i+2] = (1.0-a)*(x1[3*i+2]+2.0*PI) + a*x2[3*i+2];
    }
    
    if (v[3*i+2] > PI)
      v[3*i+2] -= 2.0*PI;

  }

  return v;

}



vector Model2DRigidMulti::StateToConfiguration(const vector &x) 
{
  return x;
}




// *********************************************************************
// *********************************************************************
// CLASS:     Model2DRigidChain
//  
// *********************************************************************
// *********************************************************************

Model2DRigidChain::Model2DRigidChain(string path = ""):Model2DRigid(path) {

  int i;

  file_istream fin(FilePath + "NumBodies");
  fin >> NumBodies;
  StopAngle = PI/1.5;

  StateDim = NumBodies+2;
  InputDim = NumBodies+2;

  file_istream fin2(FilePath + "A");
  fin2 >> A;
  
  LowerState = vector(NumBodies+2);
  UpperState = vector(NumBodies+2);

  LowerState[0] = 0.0; UpperState[0] = 100.0;
  LowerState[1] = 0.0; UpperState[1] = 100.0;

  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState[0] = 0.0; 
    LowerState[1] = 0.0; 
    for (i = 0; i < NumBodies; i++) {
      LowerState[i+2] = -StopAngle;
    }
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState[0] = 100.0; 
    UpperState[1] = 100.0; 
    for (i = 0; i < NumBodies; i++) {
      UpperState[i+2] = StopAngle;
    }
  }

  Inputs.clear();  // Otherwise its parent constructor will make some inputs
  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    // NEED A DEFAULT HERE!!!!
  }

}



vector Model2DRigidChain::LinearInterpolate(const vector &x1, 
					    const vector &x2, 
					    const double &a) {
  return (1.0-a)*x1 + a*x2;
}



vector Model2DRigidChain::StateToConfiguration(const vector &x) {
  vector q;
  int i;
  double lx,ly,ltheta;

  q = vector(3*NumBodies);
  q[0] = x[0];
  q[1] = x[1];
  q[2] = x[2];

  for (i = 1; i < NumBodies; i++) {
    lx = q[3*(i-1)];
    ly = q[3*(i-1)+1];
    ltheta = q[3*(i-1)+2];

    q[3*i] = cos(ltheta)*A[i-1]+lx;
    q[3*i+1] = sin(ltheta)*A[i-1]+ly;
    q[3*i+2] = atan2(cos(ltheta)*sin(x[2+i])+sin(ltheta)*cos(x[2+i]),
		     cos(ltheta)*cos(x[2+i])-sin(ltheta)*sin(x[2+i]));
    if (q[3*i+2] < 0.0)
      q[3*i+2] += 2.0*PI; // Force the orientation into [0,2pi)
  }

  return q;
}




vector Model2DRigidChain::StateTransitionEquation(const vector &x, const vector &u) {

  vector dx(StateDim);

  dx = u;
  return dx;
}



double Model2DRigidChain::Metric(const vector &x1, const vector &x2) {

  double rho;
  vector dtheta(StateDim);
  int i;
  
  rho = sqr(x1[0] - x2[0]) + sqr(x1[1] - x2[1]);

  for (i = 2; i < StateDim; i++) {
    dtheta[i] = min(fabs(x1[i]-x2[i]),2.0*PI - fabs(x1[i]-x2[i]));
    rho += sqr(50.0/PI*dtheta[i]);
  }

  rho = sqrt(rho);

  return rho;
}


// Make sure the joint position and limits are respected
bool Model2DRigidChain::Satisfied(const vector &x) 
{
  int i;

  for (i = 0; i < StateDim; i++)
    if ((x[i] > UpperState[i]) || (x[i] < LowerState[i]))
      return false;

  return true;
}

#include <LEDA/UNDEFINE_NAMES.h>

