//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#include <fstream.h>
#include <math.h>

#include "modelmisc.h"
#include "defs.h"

#include <LEDA/REDEFINE_NAMES.h>


// *********************************************************************
// *********************************************************************
// CLASS:     Model1D
// 
// *********************************************************************
// *********************************************************************

Model1D::Model1D(string path = ""):Model(path) {

  vector v;

  Force = 1.0;

  StateDim = 1;
  InputDim = 1;
  

  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState = vector(-50.0,-10.0);
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState = vector(50.0,10.0);
  }

  // Make the list of 1D Inputs
  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    Inputs.clear();  // Just to be safe
    v = vector(1); v[0] = 0;  // Float
    Inputs.push_back(v);
    v = vector(1); v[0] = 1;  // Accel in +x
    Inputs.push_back(v);
    v = vector(1); v[0] = -1; // Accel in -x
    Inputs.push_back(v);
  }

}



vector Model1D::Integrate(const vector &x, const vector &u, const double &h)
{
  return RungeKuttaIntegrate(x,u,h);
}


vector Model1D::StateTransitionEquation(const vector &x, const vector &u)
{
  vector dx(2);

  dx[0] = x[1];
  dx[1] = u[0]*Force;

  return dx;
}


point Model1D::StateToLedaPoint(const vector &x) {
  return point(x[0]+50.0,x[1]*(100.0/(UpperState[1]-LowerState[1]))+50.0);
}


vector Model1D::StateToConfiguration(const vector &x) {
  return vector(x[0]+50,x[1]*(100.0/(UpperState[1]-LowerState[1]))+50.0,0.0); // Yield a zero rotation every time
}

double Model1D::Metric(const vector &x1, const vector &x2) {
  double d = 0.0;
  int i;

  for (i = 0; i < 2; i++) {
    d += sqr(x1[i] - x2[i]) / (UpperState[i] - LowerState[i]);
  }
  
  //cout << "x1: " << x1 << "  x2: " << x2 << "   Metric: " << d << "\n";

  return d;
}



// *********************************************************************
// *********************************************************************
// CLASS:     ModelLinear
// 
// *********************************************************************
// *********************************************************************

ModelLinear::ModelLinear(string path = ""):Model(path) {

  vector v;
  int i;

  file_istream fin(FilePath + "A");
  fin >> A;
  file_istream fin2(FilePath + "B");
  fin2 >> B;

  StateDim = A.dim1();
  InputDim = B.dim2();

  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState = vector(StateDim);
    for (i = 0; i < StateDim; i++) {
      LowerState[i] -= 10.0;
    }
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState = vector(StateDim);
    for (i = 0; i < StateDim; i++) {
      UpperState[i] += 10.0;
    }
  }

  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    Inputs.clear();  // Just to be safe
    v = vector(InputDim);  // Initializes to the zero vector
    Inputs.push(v);
    for (i = 0; i < StateDim; i++) {
      v = vector(InputDim);
      v[i] = 1.0;
      Inputs.push(v);
      v[i] = -1.0;
      Inputs.push(v);
    }
  }

}


point ModelLinear::StateToLedaPoint(const vector &x) {
  return point(5.0*(x[0]+10.0),5.0*(x[1]+10.0));
}

vector ModelLinear::StateToConfiguration(const vector &x) {
  return vector(5.0*(x[0]+10.0),5.0*(x[1]+10.0),0.0); 
}


vector ModelLinear::Integrate(const vector &x, const vector &u, const double &h)
{
  return RungeKuttaIntegrate(x,u,h);
}


vector ModelLinear::StateTransitionEquation(const vector &x, const vector &u)
{
  vector dx(StateDim);

  dx = A * x + B * u; // This looks too easy!

  return dx;
}



// *********************************************************************
// *********************************************************************
// CLASS:     ModelND
// 
// *********************************************************************
// *********************************************************************

ModelND::ModelND(string path = ""):Model(path) {

  vector v;
  int i;

  if ((path.length() > 0)&&(path[path.length()-1] != '/'))
    path += "/";

  FilePath = path;

  if (is_file(FilePath+"Dimension")) {
    file_istream fin(FilePath + "Dimension");
    fin >> StateDim; }
  else 
    StateDim = 2;

  if (is_file(FilePath+"CorridorWidth")) {
    file_istream fin(FilePath + "CorridorWidth");
    fin >> CorridorWidth; }
  else 
    CorridorWidth = 6.0*StateDim;

  InputDim = StateDim;

  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState = vector(StateDim);
    for (i = 0; i < StateDim; i++) {
      LowerState[i] = 0.0;
    }
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState = vector(StateDim);
    for (i = 0; i < StateDim; i++) {
      UpperState[i] = 100.0;
    }
  }

  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    Inputs.clear();  // Just to be safe
    v = vector(InputDim);  // Initializes to the zero vector
    Inputs.push(v);
    for (i = 0; i < StateDim; i++) {
      v = vector(InputDim);
      v[i] = 1.0;
      Inputs.push(v);
      v[i] = -1.0;
      Inputs.push(v);
    }
  }

}


point ModelND::StateToLedaPoint(const vector &x) {
  return point(x[0],x[1]);
}


vector ModelND::StateToConfiguration(const vector &x) {
  return vector(x[0],x[1],0.0);
}


vector ModelND::StateTransitionEquation(const vector &x, const vector &u)
{
  vector dx(StateDim);

  dx = u;

  return dx;
}


vector ModelND::Integrate(const vector &x, const vector &u, const double &h)
{
  return EulerIntegrate(x,u,h);
}


// This makes a narrow circular corridor
/*
bool ModelND::Satisfied(const vector &x) {
  bool sat;
  vector sx,cx;  // cx is the center
  int i;

  sx = vector(x.dim()-1);  // a vector of the last n-1 components
  cx = vector(x.dim()-1);
  for (i = 0; i < sx.dim(); i++) {
    sx[i] = x[i+1];
    cx[i] = (UpperState[i+1] - LowerState[i+1])/2.0;
  }
  sat = ((x[0] < 30.0) || (x[0] > 70.0) || 
	 ((cx - sx).length() < CorridorWidth));

  return sat;
}
*/

// This makes a narrow square corridor
/*
bool ModelND::Satisfied(const vector &x) {
  int i;
  double maxdiff;

  if ((x[0] < 30)||(x[0] > 70.0))
    return true;

  maxdiff = 0.0;
  for (i = 1; i < x.dim(); i++) {
    if (fabs(x[i]-50.0) > maxdiff)
      maxdiff = fabs(x[i]-50.0);
  }

  return (maxdiff < CorridorWidth);
}
*/

/*
// This makes a narrow square corridor with two bends
bool ModelND::Satisfied(const vector &x) {
  int i;

  if ((x[0] < 30)||(x[0] > 70.0))
    return true;

  // Check the proper slice for last n-2 coords
  if (x.dim() > 2) {
    for (i = 2; i < x.dim(); i++) {
      if (fabs(x[i]-50.0) > CorridorWidth)
	return false;
    }
  }

  // Check upper left corridor piece
  if ((x[0] < 50) && (fabs(x[1]-70.0) < CorridorWidth))
    return true;

  // Check lower right corridor piece
  if ((x[0] >= 50)&&(fabs(x[1]-30.0) < CorridorWidth))
    return true;
  
  // Final check for center piece
  if ((x[0] < 50.0 + CorridorWidth)&&
      (x[0] > 50.0 - CorridorWidth)&&
      (x[1] < 70.0 + CorridorWidth)&&
      (x[1] > 30.0 - CorridorWidth)) 
    return true;

  // Finally, the point is in collision
  return false;
}

*/

// *********************************************************************
// *********************************************************************
// CLASS:     ModelNintegrator
// 
// *********************************************************************
// *********************************************************************


ModelNintegrator::ModelNintegrator(string path = ""):Model(path) {

  vector v;
  int i;
  double alpha,beta;

  UBound = 0.5;
  VBound = 0.5;

  StateDim = 3;
  InputDim = 2;

  if (is_file(FilePath+"LowerState"))
    ReadLowerState();
  else {
    LowerState = vector(StateDim);
    for (i = 0; i < StateDim; i++) {
      LowerState[i] -= 20.0;
    }
  }

  if (is_file(FilePath+"UpperState"))
    ReadUpperState();
  else {
    UpperState = vector(StateDim);
    for (i = 0; i < StateDim; i++) {
      UpperState[i] += 20.0;
    }
  }

  if (is_file(FilePath+"Inputs"))
    ReadInputs();
  else {
    Inputs.clear();  // Just to be safe
    for (alpha = -UBound; alpha <= UBound; 
	 alpha += 2.0*UBound/4.0) {
      for (beta = -VBound; beta <= VBound; 
	   beta += 2.0*VBound/4.0) {
	Inputs.push_back(vector(alpha,beta));
	Inputs.push_back(vector(alpha,beta)); 
      }
    }
    
  }
}



point ModelNintegrator::StateToLedaPoint(const vector &x) {
  return point(2.5*(x[0]+20.0),2.5*(x[1]+20.0));
}

vector ModelNintegrator::StateToConfiguration(const vector &x) {
  return vector(2.5*(x[0]+20.0),2.5*(x[1]+20.0),0.0); 
}


vector ModelNintegrator::Integrate(const vector &x, const vector &u, const double &h)
{
  return RungeKuttaIntegrate(x,u,h);
}


vector ModelNintegrator::StateTransitionEquation(const vector &x, const vector &u)
{
  vector dx(StateDim);

  dx[0] = u[0];
  dx[1] = u[1];
  dx[2] = x[0]*u[1] - x[1]*u[0];

  return dx;
}


