//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#include <math.h>
#include <stdio.h>

#include "planner.h"
#include "defs.h"

#include <LEDA/REDEFINE_NAMES.h>


// *********************************************************************
// *********************************************************************
// CLASS:     Planner base class
// 
// *********************************************************************
// *********************************************************************

Planner::Planner(Problem *problem):Solver(problem) {
  Reset();
}


void Planner::Reset() {
  int i;

  NumNodes = 1000;

  if (is_file(P->FilePath+"PlannerDeltaT")) {
    file_istream fin(P->FilePath + "PlannerDeltaT");
    fin >> PlannerDeltaT;
    fin.close();
  }
  else {
    PlannerDeltaT = 1.0;
  }

  if (is_file(P->FilePath+"GapError")) {
    file_istream fin(P->FilePath + "GapError");
    fin >> GapError;
    fin.close();
  }
  else {
    GapError = vector(P->StateDim);
    for (i = 0; i < P->StateDim; i++)
      GapError[i] = 1.0;
  }

  Path.clear();
  Policy.clear();

  Holonomic = is_file(P->FilePath+"Holonomic"); // Nonholonomic by default

  G.clear();
  G2.clear();
}


vector Planner::RandomState() {
  int i;
  double r;
  vector rx;

  rx = P->LowerState;
  for (i = 0; i < P->StateDim; i++) {
      R >> r; 
      rx[i] += r * (P->UpperState[i] - P->LowerState[i]);
    }

  return rx;
}



vector Planner::NormalState(vector mean, double sd = 0.5) {
  int i,j;
  double r;
  vector rx;
  bool success = false;

  rx = mean;
  for (i = 0; i < P->StateDim; i++) {
    success = false;
    while (!success) {
      rx[i] = 0.0;
      for (j = 0; j < 12; j++) {  // Increase 12 here and below for more accuracy
	R >> r; rx[i] += r;
      }
      rx[i] = (rx[i] - 12/2)*sd*(P->UpperState[i]-P->LowerState[i])+mean[i];
      if ((rx[i] <= P->UpperState[i])&&(rx[i] >= P->LowerState[i]))
	success = true;
    }
  }

  return rx;
}




void Planner::WriteGraphs(const string &fname)
{
  ofstream fout(fname);
  G.write(fout);
  G2.write(fout);
  fout.close();
}



void Planner::ReadGraphs(const string &fname)
{
  ifstream fin(fname);
  G.read(fin);
  G2.read(fin);
  fin.close();
}



void Planner::RecordSolution(const list<node> &glist, 
			     const list<node> &g2list)
{
  node n,nfirst,nlast;

  Path.clear();
  Policy.clear();

  nfirst = glist.head();
  forall(n,glist) {
      Path.push_back(G.inf(n));
      if (n != nfirst) {
	Policy.push_back(G.inf(G.first_in_edge(n)));
      }
  }

  // The GapState is always comes from last node in glist
  GapState = Path.tail();
  // Make a dummy input to get from GapState to the next state
  Policy.push_back(P->GetInputs().head());

  if (g2list.length() == 0) {
    // Push the goal state onto the end (jumps the gap)
    Path.push_back(P->GoalState);
  }
  else { // Using two graphs
    nlast = g2list.tail();
    forall(n,g2list) {
      Path.push_back(G2.inf(n));
      if (n != nlast) {
	Policy.push_back(G2.inf(G2.first_in_edge(n)));
      }
    }
  }

  //cout << "Path: " << Path << "\n";
  //cout << "Policy: " << Policy << "\n";
}



void Planner::RecordSolution(const list<node> &glist)
{
  list<node> emptylist;
  emptylist.clear(); // Make sure it is clear

  RecordSolution(glist,emptylist);
}



bool Planner::GapSatisfied(const vector &x1, const vector &x2) {
  vector x;
  int i;

  x = P->StateDifference(x1,x2);
  for (i = 0; i < P->StateDim; i++) {
    if (fabs(x[i]) > GapError[i])
      return false;
  }

  return true;
}


