//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#ifndef MSL_PLANNER_H
#define MSL_PLANNER_H

#include <LEDA/file.h>
#include <LEDA/graph.h>
#include <LEDA/list.h>
#include <LEDA/polygon.h>
#include <LEDA/vector.h>

#include "solver.h"

#include <LEDA/REDEFINE_NAMES.h>

//! The base class for all path planners

class Planner: public Solver {
 protected:
  random_source R;

  //! Choose a state at random
  vector RandomState();

  //! Pick a state using a Normal distribution
  vector NormalState(vector mean, double sd);

 public:
  //! The solution path, as a list of states
  list<vector> Path;

  //! The solution policy, as a list of inputs
  list<vector> Policy;

  //! The last state in a path before a jump occurs
  vector GapState;

  //! Set to true to ignore inputs and avoid integration (default false).
  //! This will make "regular" path planning much faster.
  bool Holonomic;

  //! How much gap error is allowed for each element in bidirectional search
  vector GapError; 

  //! A search graph
  GRAPH<vector,vector> G;  

  //! A second search graph (if needed)
  GRAPH<vector,vector> G2;
  
  //! The times associated with a solution path
  list<double> TimeList;

  //! The states associated with a solution path
  list<vector> StateList;

  //! The inputs associated with a solution path
  list<vector> InputList;

  //! Number of nodes to generate in a single execution of Plan or Construct
  int NumNodes;
  
  //! Time step to use for incremental planners
  double PlannerDeltaT;

  //! A constructor that initializes data members.
  Planner(Problem *problem);

  //! Empty destructor
  virtual ~Planner() {};

  //! Reset the planner
  virtual void Reset();

  //! Generate a planning graph
  virtual void Construct() = 0;

  //! Attempt to solve an Initial-Goal query
  virtual bool Plan() = 0;

  //! Convert a path in the graph to Path and Policy
  void RecordSolution(const list<node> &glist,
		      const list<node> &g2list);

  void RecordSolution(const list<node> &glist);

  //! Write G and G2 to a file
  virtual void WriteGraphs(const string &fname);

  //! Read G and G2 from a file
  virtual void ReadGraphs(const string &fname);

  //! Determine if the gap error is staisfied
  bool GapSatisfied(const vector &x1, const vector &x2);
};



#include <LEDA/UNDEFINE_NAMES.h>

#endif



