//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#include <math.h>
#include <stdio.h>

#include <LEDA/graph_alg.h>

#include "prm.h"
#include "defs.h"

#include <LEDA/REDEFINE_NAMES.h>


// These are used for quasi-random sampling
static int Primes[] = {2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
		       61,67,71,73,79,83,89,97,101,103,107,109,113,127,
		       131,137,139,149,151,157,163,167,173};


// *********************************************************************
// *********************************************************************
// CLASS:     PRM base class
// 
// *********************************************************************
// *********************************************************************

PRM::PRM(Problem *problem): Planner(problem) {
  MaxEdgesPerNode = 20;
  MaxNeighbors = 20;

  if (is_file(P->FilePath+"Radius")) {
    file_istream fin(P->FilePath + "Radius");
    fin >> Radius; }
  else 
    Radius = 20.0;

  SatisfiedCount = 0;
  QuasiRandom = is_file(P->FilePath + "QuasiRandom");

  //! Choose Hammersley (which is better) or Halton sequence for quasi-random points
  QuasiRandomHammersley = true;

  if (!Holonomic)
    cout << "WARNING: Differential constraints will be ignored.\n";
}




list<node> PRM::NeighboringNodes(const vector &x) {
  double d;
  node n;
  list<node> best_list;
  int k;

  best_list.clear();
  k = 0;
  forall_rev_nodes(n,G) {  // Very important to use reverse here!
    d = P->Metric(G[n],x);
    if ((d < Radius)&&(k < MaxNeighbors)) {
      best_list.push(n);
      k++;  // Count the number of attempted connections
    }
  } 
  
  //cout << "Number of Neighbors: " << best_list.length() << "\n";

  return best_list;
}




// This function essentially iterates Extend until it has to stop

bool PRM::Connect(const vector &x1, const vector &x2, vector &u_best) {
  double a;
  int k;
  vector x;

  k = (int) (P->Metric(x1,x2) / StepSize);

  //cout << "metric: " << P->Metric(x1,x2) << 
  //  "    stepsize: " << StepSize << "\n";

  if (k == 0)
    return true;

  for (a = 1.0/(k+1); a < 1.0; a += 1.0/(k+1) ) {
    x = P->InterpolateState(x1,x2,a);
    SatisfiedCount++;
    if (!P->Satisfied(x))
      return false;
  }

  return true;
}



void PRM::Construct()
{
  int i,c,k;
  vector u_best,nx;
  node n,nn;
  list<node> nhbrs;
  
  // Set the step size
  StepSize = P->Metric(P->InitialState,P->Integrate(P->InitialState,
 	     P->GetInputs(P->InitialState).head(),PlannerDeltaT));

  i = 0;
  while (i < NumNodes) {
    nx = ChooseState(i,NumNodes,P->InitialState.dim());
    SatisfiedCount++;
    i++;
    while (!P->Satisfied(nx)) {
      nx = ChooseState(i,NumNodes,P->InitialState.dim());  
      SatisfiedCount++;
      i++;
      // Keep trying until a good sample is found
    }
    nhbrs = NeighboringNodes(nx);
    nn = G.new_node(nx);

    k = 0;
    while ((k < MaxEdgesPerNode)&&(nhbrs.length() > 0)) {
      n = nhbrs.pop();
      if (Connect(G.inf(n),nx,u_best)) {
	G.new_edge(nn,n,u_best);
	G.new_edge(n,nn,u_best);
	k++;
      }
    }

    if (G.number_of_nodes() % 1000 == 0)
      cout << G.number_of_nodes() << " nodes in the PRM.\n";
  }

  node_array<int> labels(G);
  c = COMPONENTS(G,labels); // Compute connected components

  /* Show component sizes
  int count;
  for (i = 0; i < c; i++) {
    count = 0;
    forall_nodes(n,G) {
      if (labels[n] == i)
	count++;
    }
    cout << "Component " << i << " has " << count << " nodes.\n";
  }
  */

  cout << "PRM Nodes: " << G.number_of_nodes() << 
    "  Edges: " << G.number_of_edges() <<
    "  ColDet: " << SatisfiedCount << 
    "  Components: " << c << "\n";

}



bool PRM::Plan()
{    
  list<node> nhbrs;
  node n,ni,ng;
  vector u_best;
  bool success;
  list<node> path;

  // Set the step size
  StepSize = P->Metric(P->InitialState,P->Integrate(P->InitialState,
 	     P->GetInputs(P->InitialState).head(),PlannerDeltaT));

  n = ni = ng = G.choose_node();

  // Connect to the initial state
  nhbrs = NeighboringNodes(P->InitialState);
  if (nhbrs.length() > 0)
    n = nhbrs.pop();
  else {
    cout << "No neighboring nodes to the Initial State\n";
    //cout << "  FAilure\n";
    return false;
  }

  success = Connect(P->InitialState,G.inf(n),u_best);
  while((nhbrs.length() > 0)&&
	(!success)) {
    n = nhbrs.pop();
    success = Connect(P->InitialState,G.inf(n),u_best);
  }

  if (!success) {
    cout << "Failure to connect to Initial State\n";
    //cout << "  FaIlure\n";
    return false;
  }
  //else
  //  cout << "Connection to Initial State\n";

  ni = n;

  // Connect to the goal state
  nhbrs = NeighboringNodes(P->GoalState);
  if (nhbrs.length() > 0)
    n = nhbrs.pop();
  else {
    cout << "No neighboring nodes to the Goal State\n";
    //cout << "  FaiLure\n";
    return false;
  }
  
  success = Connect(P->GoalState,G.inf(n),u_best);
  while((nhbrs.length() > 0)&&
	(!success)) {
    n = nhbrs.pop();
    success = Connect(P->GoalState,G.inf(n),u_best);
  }

  if (!success) {
    cout << "Failure to connect to Goal State\n";
    //cout << "  FailUre\n";
    return false;
  }
  //else
  //  cout << "Connection to Goal State\n";

  ng = n;

  // Search G for a path from ni to ng
  if (ni == ng) {
    cout << "Initial and Goal are in the same neighborhood.\n";
    path.append(ni);
    path.append(ng);
    //cout << "  Success\n";
    RecordSolution(path);
    return true;
  }

  // Perform the search

  EdgeCost = edge_array<double>(G,1.0);
  CostToGo = node_array<double>(G);
  NextEdge = node_array<edge>(G);
  SHORTEST_PATH_T(G,ng,EdgeCost,CostToGo,NextEdge);

  if ((CostToGo[ni] > 0)&&(ni != ng)) {
    //cout << "Cost to goal node: " << CostToGo[ni] << "\n";

    // Get the path
    n = ni;
    while (n != ng) {
      path.append(n);
      n = G.opposite(n,NextEdge[n]);
    }
    path.append(ng);
  }
  else {
    cout << "  FailuRe\n";
    return false;
  }

  RecordSolution(path);
  cout << "  Success\n";

  return true;
}



vector PRM::ChooseState(int i, int maxnum, int dim) {
  if (QuasiRandom) {
    if (QuasiRandomHammersley)
      return QuasiRandomStateHammersley(i,maxnum,dim);
    else
      return QuasiRandomStateHalton(i,dim);
  }
  else
    return RandomState();
}




vector PRM::QuasiRandomStateHammersley(int i, int maxnum, int dim) {
  int j,k,r,ppow;
  vector qrx;

  if (dim > 30) {
    cout << "ERROR: Dimension too high for quasi-random samples\n";
    exit(-1);
  }

  qrx = vector(dim);

  k = i;
  
  qrx[0] = ((double) k / maxnum) * (P->UpperState[0] - P->LowerState[0]);
  qrx[0] += P->LowerState[0];
  for (j = 1; j < dim; j++) {
    qrx[j] = 0.0;
    ppow = Primes[j-1];
    k = i;
    while (k != 0) {
      r = k % Primes[j-1];
      k = (int) (k / Primes[j-1]);
      qrx[j] += (double) r / ppow;
      ppow *= Primes[j-1];
    }
    qrx[j] *= (P->UpperState[j] - P->LowerState[j]);
    qrx[j] += P->LowerState[j];
  }

  return qrx;
}


vector PRM::QuasiRandomStateHalton(int i, int dim) {
  int j,k,r,ppow;
  vector qrx;

  if (dim > 30) {
    cout << "ERROR: Dimension too high for quasi-random samples\n";
    exit(-1);
  }

  qrx = vector(dim);
  k = i;

  for (j = 0; j < dim; j++) {
    qrx[j] = 0.0;
    ppow = Primes[j];
    k = i;
    while (k != 0) {
      r = k % Primes[j];
      k = (int) (k / Primes[j]);
      qrx[j] += (double) r / ppow;
      ppow *= Primes[j];
    }
    qrx[j] *= (P->UpperState[j] - P->LowerState[j]);
    qrx[j] += P->LowerState[j];
  }

  return qrx;
}

