//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#include <fstream.h>
#include <math.h>

#include "problem.h"
#include "defs.h"

#include <LEDA/REDEFINE_NAMES.h>

// Constructor
Problem::Problem(Geom *geom, Model *model, string path = "") {

  SetGeom(geom);
  SetModel(model);

  if ((path.length() > 0)&&(path[path.length()-1] != '/'))
    path += "/";

  FilePath = path;

  if (is_file(FilePath+"InitialState"))
    ReadInitialState();
  else
    InitialState = M->LowerState + 
                   0.5*(M->UpperState - M->LowerState); 

  if (is_file(FilePath+"GoalState"))
    ReadGoalState();
  else
    GoalState = M->LowerState;

  StateDim = M->StateDim;
  InputDim = M->InputDim;
  LowerState = M->LowerState;
  UpperState = M->UpperState;

  NumBodies = G->NumBodies;
  MaxDeviates = G->MaxDeviates;
}


void Problem::SetGeom(Geom *geom) {
  G = geom;
  NumBodies = G->NumBodies;
  MaxDeviates = G->MaxDeviates;
  GeomDim = G->GeomDim;
}


void Problem::SetModel(Model *model) {
  M = model;
  StateDim = M->StateDim;
  InputDim = M->InputDim;
  LowerState = M->LowerState;
  UpperState = M->UpperState;
  if (is_file(FilePath+"InitialState"))
    ReadInitialState();
  else
    InitialState = M->LowerState + 
                   0.5*(M->UpperState - M->LowerState); 

  if (is_file(FilePath+"GoalState"))
    ReadGoalState();
  else
    GoalState = M->LowerState;
}


void Problem::ReadInitialState() {
  file_istream fin(FilePath + "InitialState");
  fin >> InitialState;
}

void Problem::ReadGoalState() {
  file_istream fin(FilePath + "GoalState");
  fin >> GoalState;
}


// In the base class, steal the following methods from Model

list<vector> Problem::GetInputs(const vector &x) {
  return M->GetInputs(x);
}

list<vector> Problem::GetInputs() {
  vector x(StateDim);

  return M->GetInputs(x);
}


vector Problem::InterpolateState(const vector &x1, const vector &x2, 
				 const double &a) {
  return M->LinearInterpolate(x1,x2,a);
}


point Problem::StateToLedaPoint(const vector &x) {
  return M->StateToLedaPoint(x);
}

// By default, don't change anything
vector Problem::StateToConfiguration(const vector &x) {
  return M->StateToConfiguration(x);
}

// Default metric: use the metric from the model
double Problem::Metric(const vector &x1, const vector &x2) {
  return M->Metric(x1,x2);
}

vector Problem::StateDifference(const vector &x1, 
				const vector &x2) {
  return M->StateDifference(x1,x2);
}

bool Problem::Satisfied(const vector &x) {
  return ((G->CollisionFree(StateToConfiguration(x)))&&
	  (M->Satisfied(x)));
}

vector Problem::Integrate(const vector &x, const vector &u, 
			  const double &deltat) {
  return M->Integrate(x,u,deltat);
} 

// In the base class, steal the following methods from Geom

bool Problem::CollisionFree(const vector &q) {
  return G->CollisionFree(q);
}


double Problem::DistanceComp(const vector &q) {
  return G->DistanceComp(q);
}

list<polygon> Problem::EnvironmentToLedaPolygons() {
  return G->EnvironmentToLedaPolygons();
}

list<polygon> Problem::RobotToLedaPolygons(const vector &q) {
  return G->RobotToLedaPolygons(q);
}

vector Problem::ConfigurationDifference(const vector &q1, 
				    const vector &q2) {
  return G->ConfigurationDifference(q1,q2);
}

#include <LEDA/UNDEFINE_NAMES.h>
