//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#include <math.h>
#include <stdlib.h>

#include "render.h"
#include "defs.h"

// Include the VCR buttons
#include "pixmaps/ff.xpm"
#include "pixmaps/pause.xpm"
#include "pixmaps/play.xpm"
#include "pixmaps/rw.xpm"
#include "pixmaps/stop.xpm"
#include "pixmaps/prev.xpm"
#include "pixmaps/next.xpm"

// Include the msl icon
#include "pixmaps/msl.xpm"


#include <LEDA/REDEFINE_NAMES.h>

////////////////////////////////////////////////////////
//
// Render Class
//
///////////////////////////////////////////////////////

Render::Render() {
  Render("");
}


Render::Render(string filepath)
{
  FilePath = filepath;
}


Render::Render(Scene *s, string filepath)
{
  SetScene(s);

  FilePath = filepath;
  ControlFreak = false;
}


// This sets up the file names for the environment and movable bodies
void Render::Init()
{
  int i;

  Reset();

  FrameTime = 0.1; // This is default; each renderer should determine this
  AnimationStartPause = 0.0; // should be in seconds
  AnimationEndPause = 0.0;
  RenderCtlWindowOn = false;

  // Colors for objects (add more if you like, but increase RENDERCOLORS)
  //RGBRed[0] = .467; RGBGreen[0] = 0.53; RGBBlue[0] = 0.6; // Light Slate Gray
  RGBRed[0] = 1.0; RGBGreen[0] = 0.0; RGBBlue[0] = 1.0; // Magenta
  RGBRed[1] = 0.8; RGBGreen[1] = 0.0; RGBBlue[1] = 0.0; // Red
  RGBRed[2] = 0.0; RGBGreen[2] = 0.7; RGBBlue[2] = 0.0; // Green
  RGBRed[3] = 0.7; RGBGreen[3] = 0.7; RGBBlue[3] = 0.0; // Yellow
  RGBRed[4] = 0.0; RGBGreen[4] = 0.7; RGBBlue[4] = 0.7; // Lightblue
  RGBRed[5] = 1.0; RGBGreen[5] = 0.0; RGBBlue[5] = 1.0; // Magenta
  RGBRed[6] = 0.0; RGBGreen[6] = 0.0; RGBBlue[6] = 0.7; // Blue
  RGBRed[7] = 1.0; RGBGreen[7] = 0.65; RGBBlue[7] = 0.0; // Orange
  RGBRed[8] = 1.0; RGBGreen[8] = 0.078; RGBBlue[8] = 0.576; // DeepPink
  RGBRed[9] = 1.0; RGBGreen[9] = 0.51; RGBBlue[9] = 0.278; // Sienna1

  // Load environment
  if (is_file(FilePath+"EnvList")) {
    file_istream fin1(FilePath + "EnvList");
    fin1 >> EnvList;
  }
  else {
    EnvList.clear(); // Make sure empty
    if (is_file(FilePath+"Obst"))
      EnvList.push("Obst"); // Use the robot from Geom
  }
  
  // Load bodies
  if (is_file(FilePath+"BodyList")) {
    file_istream fin(FilePath + "BodyList");
    fin >> BodyList;
  }
  else {
    BodyList.clear(); // Make sure empty
    if (is_file(FilePath+"Robot")) {
      BodyList.push("Robot"); // Use the robot from Geom
    }
    else { // Multiple robots 
      for (i = 0; i < S->NumBodies; i++) {
	if (is_file(string("%sRobot%d",FilePath,i))) {
	  BodyList.push_back(string("Robot%d",i)); // Use the robot from World
	}
      }
    }
  }
}



void Render::SetScene(Scene *s) 
{
  S = s; // S is the Scene for Render
}


void Render::MakeAnimationFrames(const list<vector> &xlist, double deltat) 
{
  double t;
  int i;

  StateList = xlist;
  TimeList.clear();
  
  t = 0.0;
  for (i = 0; i < xlist.length(); i++) {
    TimeList.push_back(t);
    t += deltat;
  }
  
  SetFrameList();
}


void Render::MakeAnimationFrames(const list<vector> &xlist,
				 const list<double> &timelist) 
{
  StateList = xlist;
  TimeList = timelist;
  SetFrameList();
}



void Render::SetFrameList()
{
  double ltime; // Lower time
  double utime; // Upper time
  double ctime; // Current time
  double lasttime;
  double lambda;
  vector sc; // scene configuration
  vector lx,ux; // lower and upper states
  int i;
  list_item li,tli,stli,tstli;

  //cout << "Running SetFrameList\n";
  //cout << "StateList: " << StateList << "\n";
  //cout << "TimeList: " << TimeList << "\n";

  FrameList.clear();
  NumFrames = 0;

  if (TimeList.length() < 2) {
    if (TimeList.length() == 1) {
      FrameList.push(S->StateToSceneConfiguration(StateList.head()));
      NumFrames = 1;
    }
    return;
  }

  // Make the starting pause
  for (i = 0; i < (int) (AnimationStartPause/FrameTime); i++) {
    FrameList.push_back(S->StateToSceneConfiguration(StateList.head()));
    NumFrames++;
  }

  li = TimeList.succ(TimeList.first());
  stli = StateList.succ(StateList.first());
  utime = TimeList.inf(li);  // Time index of second element
  ux = StateList.inf(stli);
  ltime = TimeList.head();
  lx = StateList.head();
  ctime = ltime;
  lasttime = TimeList.tail();
  while (ctime <= lasttime) {
    while ((utime < ctime)&&  // Make sure the ctime is within upper and lower
	   (TimeList.succ(li))) {
      li = TimeList.succ(li);
      stli = StateList.succ(stli);
      utime = TimeList.inf(li);
      ux = StateList.inf(stli);
    }
    tli = TimeList.pred(li);
    ltime = TimeList.inf(tli);
    tstli = StateList.pred(stli);
    lx = StateList.inf(tstli);
    while (ltime > ctime) { // This can happen sometimes (I think)!
      tli = TimeList.pred(tli);
      tstli = StateList.pred(tstli);
      ltime = TimeList.inf(tli);
      lx = StateList.inf(tstli);
    }

    lambda = (ctime - ltime)/(utime - ltime);
    //cout << "lambda: " << lambda << "\n";

    // Interoplate based on states lx and ux
    sc = S->InterpolatedSceneConfiguration(lx,ux,lambda);
    FrameList.push_back(sc);
    NumFrames++;
    ctime += FrameTime;
  }
  
  // Add in the last frame, if necessary
  if (FrameList.tail() != (S->StateToSceneConfiguration(StateList.tail())))
    FrameList.push_back(S->StateToSceneConfiguration(StateList.tail()));

  // Make the ending pause
  for (i = 0; i < (int) (AnimationEndPause/FrameTime); i++) {
    FrameList.push_back(S->StateToSceneConfiguration(StateList.tail())); 
    NumFrames++;
  }

  //vector x;
  //cout << "StateList:\n";
  //forall(x,StateList)
  //  cout << x << "\n";
  //cout << "TimeList: " << TimeList << "\n";
  //cout << "\n\nFrameList:\n";
  //forall(x,FrameList)
  //  cout << x << "\n";
  //cout << "NumFrames: " << NumFrames << "\n";
}



void Render::SetCurrentAnimationFrame() {
  vector c(S->SceneConfigurationDim);
  int num,skip;

  // How long has the frame been stuck?
  FrameStuckTime = used_time() - LastFrameTime;

  skip = (int) (AnimationTimeScale * FrameStuckTime / FrameTime);

  if (skip > 0) {
    AnimationFrameIndex += skip;
    used_time(LastFrameTime);
  }
    
  num = FrameList.length();
  if (AnimationFrameIndex > num - 1) {
    AnimationFrameIndex = 0;
    used_time(LastFrameTime);
  }

  if (AnimationFrameIndex < 0) {
    AnimationFrameIndex = 0;
    used_time(LastFrameTime);
  }

  if (num > 0) {
    c = FrameList.contents(FrameList.get_item(AnimationFrameIndex));
  }

  CurrentAnimationFrame = c;
}



void Render::ButtonHandle(int b) 
{
  //cout << "Button " << b << "\n";

  switch (b) {

  case 70: // Play
    if (!AnimationActive) {
      used_time(LastFrameTime);
      AnimationActive = true;
    }
    break;
  case 71: // Stop
      AnimationActive = false;
      AnimationFrameIndex = 0;
      LastFrameTime = 0.0;;
      used_time(LastFrameTime);
    break;
  case 72: // Pause
    if (AnimationActive) {
      AnimationActive = false;
    }
    else {
      AnimationActive = true;
      used_time(LastFrameTime);
    }
    cout << "Frame: " << AnimationFrameIndex << "   Time stamp: " 
	 << AnimationFrameIndex*FrameTime << "s\n";
    break;
  case 73: // Deccelerate animation
    AnimationTimeScale /= 1.41421356237; // Two presses doubles the speed
    break;
  case 74: // Accelerate animation
    AnimationTimeScale *= 1.41421356237; // Two presses doubles the speed
    break;
  case 75:
    Reset();
    break;
  case 76: // Prev frame
    AnimationActive = false;
    if (AnimationFrameIndex > 0)
      AnimationFrameIndex--;
    else
      AnimationFrameIndex = FrameList.length() - 1;
    CurrentAnimationFrame = FrameList.contents(FrameList.get_item(AnimationFrameIndex));
    ShowCurrentAnimationFrame();
    cout << "Frame: " << AnimationFrameIndex << "   Time stamp: " 
	 << AnimationFrameIndex*FrameTime << "s\n";
   break;
  case 77: // Next frame
    AnimationActive = false;
    if (AnimationFrameIndex < FrameList.length() - 1)
      AnimationFrameIndex++;
    else
      AnimationFrameIndex = 0;
    CurrentAnimationFrame = FrameList.contents(FrameList.get_item(AnimationFrameIndex));
    ShowCurrentAnimationFrame();
    cout << "Frame: " << AnimationFrameIndex << "   Time stamp: " 
	 << AnimationFrameIndex*FrameTime << "s\n";
    break;
  default: cout << "Function not implemented\n";
    break;
  }
}


void Render::MainLoop(Gui *g) {

  g->Finished = false;

  while (!g->Finished) {
    g->HandleEvents();
    HandleEvents();
  }

}



void Render::Reset() {
  AnimationActive = false;
  AnimationFrameIndex = 0;
  if (FrameList.length() > 0)
    CurrentAnimationFrame = FrameList.contents(FrameList.get_item(0));
  else
    CurrentAnimationFrame = vector(S->SceneConfigurationDim);
  AnimationTimeScale = 1.0;
  LastFrameTime = 0.0;
  used_time(LastFrameTime);
  AttachedCameraOn = false;
  BoundingBoxOn = false;
  MultipleViewsOn = false;
  ShowPathOn = false;
  AmbientLight = 0.2;
}



void Render::ToggleRenderCtlWindow() 
{
  char *play,*stop,*pause,*rw,*ff,*prev,*next;
  RenderCtlWindowOn = !RenderCtlWindowOn;
  if (RenderCtlWindowOn) {
    RenderCtlWindow = new window(200,155,"MSL: Render Control Window");
    RenderCtlWindow->set_icon_pixrect(RenderCtlWindow->create_pixrect(msl_xpm));
    RenderCtlWindow->double_item("AnimationTimeScale",AnimationTimeScale);
    RenderCtlWindow->double_item("AmbientLight",AmbientLight);
    RenderCtlWindow->bool_item("BoundingBox",BoundingBoxOn);
    RenderCtlWindow->bool_item("MultipleViews",MultipleViewsOn);
    RenderCtlWindow->bool_item("AttachedCamera",AttachedCameraOn);
    RenderCtlWindow->bool_item("ShowPath",ShowPathOn);
    
    // Make the animation buttons
    play = RenderCtlWindow->create_pixrect(play_xpm);
    stop = RenderCtlWindow->create_pixrect(stop_xpm);
    pause = RenderCtlWindow->create_pixrect(pause_xpm);
    rw = RenderCtlWindow->create_pixrect(rw_xpm);
    ff = RenderCtlWindow->create_pixrect(ff_xpm);
    prev = RenderCtlWindow->create_pixrect(prev_xpm);
    next = RenderCtlWindow->create_pixrect(next_xpm);
    RenderCtlWindow->buttons_per_line(7);
    RenderCtlWindow->button(stop,stop,"",71);
    RenderCtlWindow->button(prev,prev,"",76);
    RenderCtlWindow->button(pause,pause,"",72);
    RenderCtlWindow->button(next,next,"",77);
    RenderCtlWindow->button(rw,rw,"",73);
    RenderCtlWindow->button(play,play,"",70);
    RenderCtlWindow->button(ff,ff,"",74);

    RenderCtlWindow->button("Reset",75);

    RenderCtlWindow->display();
  }
  else {
    RenderCtlWindow->close();
  }
}



#include <LEDA/UNDEFINE_NAMES.h>




