//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#include <math.h>
#include <stdlib.h>

#include "renderleda.h"
#include "defs.h"

#include <LEDA/REDEFINE_NAMES.h>


////////////////////////////////////////////////////////
//
// RenderLEDA Class
//
///////////////////////////////////////////////////////

RenderLeda::RenderLeda(): Render()
{
}


RenderLeda::RenderLeda(string filepath=""): Render(filepath)
{
}


RenderLeda::RenderLeda(Scene *s, string filepath): Render(s,filepath)
{
}


RenderLeda::~RenderLeda()
{
}


void RenderLeda::Init() {
  string fname;
  int i;
  list<polygon> plist;
  list<Triangle> trlist;

  Render::Init();

  TransformDim = 6;
  FirstFrameDisplayed = false;

  W = new window(600,600,"Motion Strategy Library, Iowa State University, 1998-2000");
  // This is one makes the video conversion easier
  //w = new window(600,450+37+50,"Rapidly-Exploring Random Trees          Steve LaValle, Iowa State, 1999");
  //w->init(-16.0,117.0,-5.0);

  W->display();
  W->set_mode(xor_mode);

  // Read the environment
  Env = array<list<polygon> >(EnvList.length());
  i = 0;
  forall(fname,EnvList) {
    file_istream fin(FilePath + fname);
    if (S->GeomDim == 3) {
      fin >> trlist;
      plist = TrianglesToPolygons(trlist); // Defined in 3Dtriangle.C
    }
    else
      fin >> plist;
    Env[i] = plist;
    plist.clear();  
    i++;
  }
  
  DrawEnv(); // Draw the environment
  
  // Read the bodies
  Bodies = array<list<polygon> >(BodyList.length());
  i = 0;
  forall(fname,BodyList) {
    file_istream fin(FilePath + fname);
    if (S->GeomDim == 3) {
      fin >> trlist;
      plist = TrianglesToPolygons(trlist); // Defined in 3Dtriangle.C
    }
    else
      fin >> plist;
    Bodies[i] = plist;
    plist.clear();
    i++;
  }
}



void RenderLeda::DrawEnv(){
  int i;
  polygon p;

  for (i = 0; i < Env.size(); i++) {
    forall(p,Env[i]) {
      W->draw_filled_polygon(p,black);
    }
  }
}



void RenderLeda::DrawBodies(const vector &x)
{
  int i;
  polygon p;

  for (i = 0; i < Bodies.size(); i++) {
    forall(p,Bodies[i]) {
      W->draw_filled_polygon(p,green);
    }
  }
}



void RenderLeda::ShowCurrentAnimationFrame()
{
  polygon p,tp;
  list<polygon> body;
  int i;
  vector sc;

  if ((!FirstFrameDisplayed)||(PreviousFrame != CurrentAnimationFrame)) {

    // Unshow the last frame
    if (FirstFrameDisplayed) {
      sc = PreviousFrame;
      for (i = 0; i < S->NumBodies; i++) {
	forall(p,Bodies[i]) {
	  tp = p.rotate(sc[i*TransformDim+TransformDim-1]);  // yaw
	  tp = tp.translate(sc[i*TransformDim],sc[i*TransformDim+1]);
	  W->draw_filled_polygon(tp,yellow);
	  W->draw_polygon(tp,black);
	}
      }
    }
    
    sc = CurrentAnimationFrame;
    for (i = 0; i < S->NumBodies; i++) {
      forall(p,Bodies[i]) {
	tp = p.rotate(sc[i*TransformDim+TransformDim-1]);  // yaw
	tp = tp.translate(sc[i*TransformDim],sc[i*TransformDim+1]);
	W->draw_filled_polygon(tp,yellow);
	W->draw_polygon(tp,black);
      }
    }
    
    PreviousFrame = sc;
    FirstFrameDisplayed = true;
  }
}



void RenderLeda::DrawPath()
{
  vector sc;
  polygon p,tp;
  int i;

  Reset();  // Reset the window

  forall(sc,FrameList) {
    //cout << "sc: " << sc << "\n";
    for (i = 0; i < S->NumBodies; i++) {
      forall(p,Bodies[i]) {
	tp = p.rotate(sc[i*TransformDim+TransformDim-1]);  // yaw
	tp = tp.translate(sc[i*TransformDim],sc[i*TransformDim+1]);
	W->draw_filled_polygon(tp,yellow);
	W->draw_polygon(tp,black);
      }
    }
  }
}



void RenderLeda::Terminate()
{
  // Terminate processes and exit
}



void RenderLeda::HandleEvents()
{
  // Handle events from render control window
  Render::HandleEvents();

  if (AnimationActive) {
    SetCurrentAnimationFrame();
    ShowCurrentAnimationFrame();
  }
}


void RenderLeda::Reset()
{
  W->clear();
  DrawEnv(); // Draw the environment
}


#include <LEDA/UNDEFINE_NAMES.h>





