//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#include <math.h>
#include <stdio.h>

#include "rrt.h"
#include "defs.h"

#include <LEDA/REDEFINE_NAMES.h>


// *********************************************************************
// *********************************************************************
// CLASS:     RRT base class
// 
// *********************************************************************
// *********************************************************************

RRT::RRT(Problem *problem): Planner(problem) {

  UseANN = false;
  NumNodes = 1000;

  if (is_file(P->FilePath+"ConnectTimeLimit")) {
    file_istream fin(P->FilePath + "ConnectTimeLimit");
    fin >> ConnectTimeLimit;
    fin.close();
  }
  else {
    ConnectTimeLimit = INFINITY; // Go "all the way" in Connect
  }

  Reset();
}



void RRT::Reset() {
  Planner::Reset();

  SatisfiedCount = 0;
  GoalDist = P->Metric(P->InitialState,P->GoalState);
  BestState = P->InitialState;

  GEdgeTime.init(G,PlannerDeltaT);
  G2EdgeTime.init(G2,PlannerDeltaT);

#ifdef USE_ANN
  MAG = MultiANN(&G);
  MAG2 = MultiANN(&G2);
#endif

}


// Return the best new state in nx_best
// success will be false if no action is collision free
vector RRT::SelectBestInput(const vector &x1, const vector &x2, 
			    vector &nx_best, bool &success,
			    bool forward = true)
{
  vector u,u_best,nx;
  double d,d_min;
  success = false;
  d_min = (forward) ? P->Metric(x1,x2) : P->Metric(x2,x1);
  list<vector> il = P->GetInputs(x1);

  if (Holonomic) { // Just do interpolation
    u_best = P->InterpolateState(x1,x2,0.1) - x1;
    u_best = u_best.norm(); // Normalize the direction
    nx_best = P->Integrate(x1,u_best,PlannerDeltaT);
    SatisfiedCount++;
    if (P->Satisfied(nx_best))
      success = true;
  }
  else {  // Nonholonomic (the more general case -- look at Inputs)
    forall(u,il) {
      if (forward)
	nx = P->Integrate(x1,u,PlannerDeltaT);
      else
	nx = P->Integrate(x1,u,-PlannerDeltaT);
      
      d  = (forward) ? P->Metric(nx,x2): P->Metric(x2,nx);
      
      SatisfiedCount++;
      
      if ((d < d_min)&&(x1 != nx)) { 
	if (P->Satisfied(nx)) {
	  d_min = d; u_best = u; nx_best = nx; success = true;
	}
      }
    }
  }

  // Uncomment below for random inputs!
  //int n = il.length();
  //R >> d;
  //u_best = il.inf(il.get_item((int) (d*n)));
  //nx_best = P->Integrate(x1,u_best,PlannerDeltaT);
    
  //cout << "u_best: " << u_best << "\n";

  return u_best;
}



vector RRT::SelectRandomInput(const vector &x1, const vector &x2, 
			    vector &nx_best, bool &success,
			    bool forward = true)
{
  vector u,u_best;
  double d;
  success = false;
  int i,n;
  
  i = 0;
  list<vector> il = P->GetInputs(x1);
  n = il.length();
  while ((i < n)&&(!success)) {
    R >> d;
    u_best = il.inf(il.get_item((int) (d*n)));
    nx_best = P->Integrate(x1,u_best,PlannerDeltaT);
    if (P->Satisfied(nx_best))
      success = true;
    i++;
  }

  return u_best;
}



node RRT::NearestNeighbor(const vector &x, const GRAPH<vector,vector> &g,
			  bool forward = true) {
  double d,d_min;
  node n,n_best;
  
  n_best = n = g.first_node(); // Keeps the warnings away
  d_min = INFINITY; d = 0.0;

#ifdef USE_ANN
  if (!UseANN) {
#endif
    forall_nodes(n,g) {
      d = (forward) ? P->Metric(g[n],x) : P->Metric(x,g[n]);
      if (d < d_min) {
	d_min = d; n_best = n; 
      }
    } 
#ifdef USE_ANN
  }
  else {
    if (&g == &G)
      n_best = MAG.NearestNeighbor(x);
    else
      n_best = MAG2.NearestNeighbor(x);
  }
#endif

  //cout << "n_best: " << n_best << "info: " << g.inf(n_best) << "\n";
  
  //n_best =  g.choose_node(); // Pick a node at random

  return n_best;
}



bool RRT::Extend(const vector &x, 
		 GRAPH<vector,vector> &g,
		 node &nn, bool forward = true) {
  node n_best;
  vector nx,u_best;
  bool success;
  
  n_best = NearestNeighbor(x,g,forward);
  u_best = SelectBestInput(g.inf(n_best),x,nx,success,forward);
  // nx gets next state
  if (success) {   // If a collision-free input was found
    nn = g.new_node(nx); // Make a new node with state nx
    g.new_edge(n_best,nn,u_best);
    //cout << "New node: " << g.inf(nn) << "   " << "u:" << u_best << "\n";
  }

  return success;
}



// This function essentially iterates Extend until it has to stop
// The same action is used for every iteration
bool RRT::Connect(const vector &x, 
		  GRAPH<vector,vector> &g,
		  node &nn, bool forward = true) {
  node nn_prev,n_best;
  vector nx,nx_prev,u_best;
  bool success;
  double d,d_prev,clock;
  int steps;
  
  n_best = NearestNeighbor(x,g,forward);
  u_best = SelectBestInput(g.inf(n_best),x,nx,success,forward); 
  steps = 0;
           // nx gets next state
  if (success) {   // If a collision-free input was found
    d = P->Metric(nx,x); d_prev = d;
    nx_prev = nx; // Initialize
    nn = n_best;
    clock = PlannerDeltaT;
    while ((P->Satisfied(nx))&&
	   (clock <= ConnectTimeLimit)&&
	   (d <= d_prev))
      {
        SatisfiedCount++;
	steps++; // Number of steps made in connecting
	nx_prev = nx;
	d_prev = d; nn_prev = nn;
	// Uncomment line below to select best action each time
	//u_best = SelectBestInput(g.inf(nn),x,nx,success,forward); 
	if (Holonomic) { 
	    nx = P->Integrate(nx_prev,u_best,PlannerDeltaT);
	}
	else { // Nonholonomic
	  if (forward)
	    nx = P->Integrate(nx_prev,u_best,PlannerDeltaT);
	  else
	    nx = P->Integrate(nx_prev,u_best,-PlannerDeltaT);
	}
	d = P->Metric(nx,x);
	clock += PlannerDeltaT;
	// Uncomment the subsequent two lines to
	//   make each intermediate node added
	//nn = g.new_node(nx_prev); // Make a new node
	//g.new_edge(nn_prev,nn,u_best);
      }
    nn = g.new_node(nx_prev); // Make a new node
    g.new_edge(n_best,nn,u_best);
    if (&g == &G)
      GEdgeTime[nn] = steps*PlannerDeltaT;
    else
      G2EdgeTime[nn] = steps*PlannerDeltaT;
  }

  return success;
}


// Like RRT::Extend, except try ALL inputs
bool RRT::Project(const vector &x, 
		 GRAPH<vector,vector> &g,
		 node &nn, bool forward = true) {
  node n_best;
  vector nx,x1,u;
  bool success;
  
  n_best = NearestNeighbor(x,g,forward);
  x1 = g.inf(n_best);
  list<vector> il = P->GetInputs(x1);
  success = false;
  forall(u,il) {
    if (forward)
      nx = P->Integrate(x1,u,PlannerDeltaT);
    else
      nx = P->Integrate(x1,u,-PlannerDeltaT); 
    if (P->Satisfied(nx)) {   // If a collision-free input was found
      nn = g.new_node(nx); // Make a new node with state nx
      g.new_edge(n_best,nn,u);
      success = true;
      //cout << "New node: " << g.inf(nn) << "   " << "u:" << u << "\n";
    }
  } 

  return success;
}



// Edges of G are input vectors, and vertices are states
// Note that this method also adds to an existing RRT
void RRT::Construct()
{
  int i;
  vector u_best,nx;
  node nn;
  double d;

  // Make the root node of G
  if (G.number_of_nodes() == 0) {
    nn = G.new_node(P->InitialState);
    GEdgeTime[nn] = 0.0;
  }

  for (i=0; i < NumNodes; i++) {
    if (Extend(ChooseState(),G,nn)) { 
      d = P->Metric(G.inf(nn),P->GoalState);
      if (d < GoalDist)  {  // Decrease if goal closer
	GoalDist = d;
	BestState = G.inf(nn);
      }
      if (G.number_of_nodes() % 1000 == 0)
	cout << G.number_of_nodes() << " nodes in the RRT.\n";
    }
  }

}


bool RRT::Plan()
{
  int i;
  double d;
  node n,nn,n_goal;
  vector nx,u_best;
  double ptime;
  list<node> path;

  // Make the root node of G
  if (G.number_of_nodes() == 0) {
    n = G.new_node(P->InitialState);
    GEdgeTime[n] = 0.0;
  }

  i = 0;
  n = NearestNeighbor(P->GoalState,G);
  n_goal = n;

  GoalDist = P->Metric(G.inf(n),P->GoalState);
  while ((i < NumNodes)&&(!GapSatisfied(G.inf(n_goal),P->GoalState))) {
    if (Extend(ChooseState(),G,nn)) { 
      d = P->Metric(G.inf(nn),P->GoalState);
      if (d < GoalDist) {  // Decrease if goal closer
	GoalDist = d;
	BestState = G.inf(nn);
	n_goal = nn;
	//cout << "GoalDist " << GoalDist << "\n";
      }
    }
    i++;
  }

  // Get the solution path
  if (GapSatisfied(G.inf(n_goal),P->GoalState)) {
    cout << "Successful Path Found\n";
    path = PathToLeaf(n_goal,G);
    // Make the correct times
    ptime = 0.0; TimeList.clear();
    forall(n,path) {
      ptime += GEdgeTime[n];
      TimeList.push_back(ptime);
    }
    RecordSolution(path); // Write to Path and Policy
    return true;
  }
  else
    return false;
}



vector RRT::ChooseState() {
  return RandomState();
}


// The algorithm actually starts at the leaf and stops at the root
list<node> RRT::PathToLeaf(const node &n, const GRAPH<vector,vector> &g) 
{
   list<node> path;
   node ni;
   int i;

   path.clear();
   path.push(n); 

   i = 0;
   ni = n;
   while ((ni != g.first_node())&&(i < g.number_of_nodes())) {
     path.push(g.source(g.first_in_edge(ni)));
     ni = g.source(g.first_in_edge(ni));
     i++;
   }

   return path;
}



// *********************************************************************
// *********************************************************************
// CLASS:     RRTGoalBias
//
// Occasionally pick the goal state instead of a random sample. 
// *********************************************************************
// *********************************************************************


RRTGoalBias::RRTGoalBias(Problem *p):RRT(p) {
  if (is_file(P->FilePath+"GoalProb")) {
    file_istream fin(P->FilePath + "GoalProb");
    fin >> GoalProb;
    fin.close();
  }
  else {
    GoalProb = 0.05;
  }
}



vector RRTGoalBias::ChooseState()
{
  double rv;

  R >> rv;
  if (rv > GoalProb)
    return RandomState();
  else
    return P->GoalState;
}


// *********************************************************************
// *********************************************************************
// CLASS:     RRTGoalPull
// 
// *********************************************************************
// *********************************************************************


RRTGoalPull::RRTGoalPull(Problem *p):RRT(p) {
}



void RRTGoalPull::Construct()
{
  int i;
  vector u_best,nx;
  node nn,nn2;
  bool success;

  // Make the root node of G
  if (G.number_of_nodes() == 0) {
    nn = G.new_node(P->InitialState);
    GEdgeTime[nn] = 0.0;
  }

  for (i=0; i < NumNodes; i++) {
    if (Extend(ChooseState(),G,nn)) { 
      u_best = SelectBestInput(G.inf(nn),P->GoalState,nx,success); // nx gets next state
      if (success) {   // If a collision-free input was found
	nn2 = G.new_node(nx); // Make a new node with state nx
	G.new_edge(nn,nn2,u_best);
      }
    }
  }
}


// *********************************************************************
// *********************************************************************
// CLASS:     RRTCon
// 
// Make the regular RRT greedy by attempting to connect all the way
// to the random sample, or go as far as possible before colliding.
// *********************************************************************
// *********************************************************************


// Build an RRT that tries to go as far as it can for each edge
RRTCon::RRTCon(Problem *p):RRTGoalBias(p) {
  if (!is_file(P->FilePath+"GoalProb")) {
    GoalProb = 0.0; // No bias by default
  }
}


void RRTCon::Construct()
{
  int i;
  vector u_best,nx,nx_prev;
  node nn;


  // Make the root node of G
  if (G.number_of_nodes() == 0) {
    nn = G.new_node(P->InitialState);
    GEdgeTime[nn] = 0.0;
  }

  for (i=0; i < NumNodes; i++) {
    Connect(ChooseState(),G,nn);
  }
}



bool RRTCon::Plan()
{
  int i;
  double d;
  node n,nn,n_goal;
  vector nx,u_best;
  double ptime;
  list<node> path;

  // Make the root node of G
  if (G.number_of_nodes() == 0) {
    n = G.new_node(P->InitialState);
    GEdgeTime[n] = 0.0;
  }

  i = 0;
  n = NearestNeighbor(P->GoalState,G);
  n_goal = n;

  GoalDist = P->Metric(G.inf(n),P->GoalState);
  while ((i < NumNodes)&&(!GapSatisfied(G.inf(n_goal),P->GoalState))) {
    if (Connect(ChooseState(),G,nn)) { 
      d = P->Metric(G.inf(nn),P->GoalState);
      if (d < GoalDist) {  // Decrease if goal closer
	GoalDist = d;
	BestState = G.inf(nn);
	n_goal = nn;
	//cout << "GoalDist " << GoalDist << "\n";
      }
    }
    i++;
  }

  // Get the solution path
  if (GapSatisfied(G.inf(n_goal),P->GoalState)) {
    cout << "Successful Path Found\n";
    path = PathToLeaf(n_goal,G);
    // Make the correct times
    ptime = 0.0; TimeList.clear();
    forall(n,path) {
      ptime += GEdgeTime[n];
      TimeList.push_back(ptime);
    }
    RecordSolution(path); // Write to Path and Policy
    return true;
  }
  else
    return false;

}


// *********************************************************************
// *********************************************************************
// CLASS:     RRTDual
// 
// The dual-tree planner used in LaValle, Kuffner, ICRA 1999.  Each 
// tree is extended toward a randomly-sampled point.  RRTExtExt is
// generally better.
// *********************************************************************
// *********************************************************************

RRTDual::RRTDual(Problem *p):RRT(p) {
}


void RRTDual::RecoverSolution(const node &n1, const node &n2) {
  list<node> path,path2;
  double ptime;
  node n;

  path = PathToLeaf(n1,G);
  // Figure out forward timings
  ptime = 0.0; TimeList.clear();
  forall(n,path) {
    ptime += GEdgeTime[n];
    TimeList.push_back(ptime);
  }
  
  path2 = PathToLeaf(n2,G2); path2.reverse();
  // Figure out backwards timings
  ptime += PlannerDeltaT; // Add a time step for the gap
  forall(n,path2)  { // Compute total time in G2 part of path
    TimeList.push_back(ptime);
    ptime += G2EdgeTime[n];
  }
  
  //cout << "Path: " << path;
  //cout << "Path2: " << path2;
  //cout << "\n TimeList: " << TimeList << " L: " << TimeList.length();

  RecordSolution(path,path2); // Write to Path and Policy
}



bool RRTDual::Plan()
{
  int i;
  vector rx,u_best,nx;
  node nn,nn2;
  bool connected;
  list<node> path1;

  connected = false;

  float t = used_time();

  //int total = 0;
  //for (int j=0; j < 100; j++) {

  //G.clear();
  if (G.number_of_nodes() == 0) {
    nn = G.new_node(P->InitialState);
    GEdgeTime[nn] = 0.0;
  }

  //G2.clear();
  if (G2.number_of_nodes() == 0) {
    nn = G2.new_node(P->GoalState);
    G2EdgeTime[nn] = 0.0;
  }

  nn = G.first_node();
  nn2 = G2.first_node();

  i = 0;
  connected = false;
  while ((i < NumNodes) && (!connected)) {
    rx = ChooseState();
    Extend(rx,G,nn);
    Extend(rx,G2,nn2,false);  // false means reverse-time integrate

    if (GapSatisfied(G.inf(nn),G2.inf(nn2))) {
      cout << "CONNECTED!!  Nodes: " << 
	G.number_of_nodes() + G2.number_of_nodes() << "\n";
      connected = true;
      RecoverSolution(nn,nn2);
    }

    i++;
  }

  if (!connected)
    cout << "Failure to connect after " << G.number_of_nodes()+G2.number_of_nodes() 
	 << " nodes\n";

  cout << "Collision Detection Calls: " << SatisfiedCount << "\n";

  //total += i;
  // }
  //cout << "Avg: " << total/100 << "\n";

  cout << "Planning Time: " << ((double)used_time(t)) << "s\n"; 

  return connected;
}





// *********************************************************************
// *********************************************************************
// CLASS:     RRTExtExt
// 
// A dual-tree planner in which computation is balanced between 
// exploring and connecting trees to each other.  Greedier verions
// of this are RRTExtCon and RRTConCon.
// *********************************************************************
// *********************************************************************

RRTExtExt::RRTExtExt(Problem *p):RRTDual(p) {
}


bool RRTExtExt::Plan()
{
  int i;
  vector u_best,nx,nx2;
  node nn,nn2;
  bool connected;
  list<node> path1;

  connected = false;

  float t = used_time();

  //int total = 0;
  //for (int j=0; j < 100; j++) {
  
  //G.clear();
  if (G.number_of_nodes() == 0) {
    nn = G.new_node(P->InitialState);
    GEdgeTime[nn] = 0.0;
  }
  //G2.clear();
  if (G2.number_of_nodes() == 0) {
    nn = G2.new_node(P->GoalState);
    G2EdgeTime[nn] = 0.0;
  }

  nn = G.first_node();
  nn2 = G2.first_node();

  i = 0;
  connected = false;
  while ((i < NumNodes) && (!connected)) {
    if (Extend(ChooseState(),G,nn)) {
      if (Extend(G.inf(nn),G2,nn2,false)) {
	i++;
	if (GapSatisfied(G.inf(nn),G2.inf(nn2))) {
	  cout << "CONNECTED!!  Nodes: " << 
	    G.number_of_nodes()+G2.number_of_nodes() 
	       << "\n";
	  connected = true;
	  RecoverSolution(nn,nn2);
	}
      }
    }
    i++;

    if ((!connected) && (Extend(ChooseState(),G2,nn,false))) {
      if (Extend(G.inf(nn),G,nn2)) {
	i++;
	if (GapSatisfied(G.inf(nn2),G2.inf(nn))) {
	  cout << "CONNECTED!!  Nodes: " << 
	    G.number_of_nodes()+G2.number_of_nodes() << "\n";
	  connected = true;
	  RecoverSolution(nn2,nn);
	}
      }
    }
  }

  i++;

  if (!connected)
    cout << "Failure to connect after " << 
      G.number_of_nodes()+G2.number_of_nodes() 
	 << " nodes\n";

  cout << "Collision Detection Calls: " << SatisfiedCount << "\n";

  //total += i;
  // }
  //cout << "Avg: " << total/100 << "\n";
  
  cout << "Planning Time: " << ((double)used_time(t)) << "s\n"; 

  return connected;
}




// *********************************************************************
// *********************************************************************
// CLASS:     RRTGoalZoom
// 
// Bias sampling towards a region that contains the goal.  The 
// region shrinks around the goal as the RRT comes nearer.  This
// planner is based on a class project at Iowa State by Jun Qu in 1999.
// *********************************************************************
// *********************************************************************

RRTGoalZoom::RRTGoalZoom(Problem *p):RRT(p) {
  GoalProb = 0.05;
  ZoomProb = 0.5;
  ZoomFactor = 2.0;
}


// Zoom using a square box
vector RRTGoalZoom::ChooseState()
{
  double rv,r,diff;
  vector zx;
  int i;

  R >> rv;
  diff = 0.0;
  zx = P->LowerState;  // Initialize to anything

  if (rv < GoalProb)
    return P->GoalState;
  else {
    if (rv < (ZoomProb+GoalProb)) {
      for (i = 0; i < P->GoalState.dim(); i++) {
	if (fabs(P->GoalState[i] - BestState[i]) > diff)
	  diff = fabs(P->GoalState[i] - BestState[i]);
      }
      for (i = 0; i < P->GoalState.dim(); i++) {
	R >> r;
	zx[i] += P->GoalState[i] - diff + 2.0*r*ZoomFactor*diff;
      }
      return zx;
    }
    else
      return RandomState();
  }
  
}


// *********************************************************************
// *********************************************************************
// CLASS:     RRTPolar
// 
// Instead of random sampling, attempt to gradually bias samples
// toward the goal.
// *********************************************************************
// *********************************************************************

RRTPolar::RRTPolar(Problem *p):RRT(p) {
  // RadiusExp = 1.0/(P->InitialState.dim() - 1);
  RadiusExp = 1.0;
}






// This implementation ignores C-space topology, VERY BAD!
vector RRTPolar::ChooseState()
{
  double r,w;
  vector zx;
  int i,j;
  bool success;
  
  w = 0.0;
  zx = P->GoalState;  // Initialize to anything

  success = false;

  while (!success) {
    for (i = 0; i < P->GoalState.dim(); i++) {
      // Generate sample from N(0,1)
      zx[i] = 0.0;
      for (j = 0; j < 12; j++) {
	R >> r; zx[i] += r;
      }
      zx[i] -= 6.0;
      w += zx[i]*zx[i];
    }
    
    w = sqrt(w);
    
    //cout << "RadiusExp: " << RadiusExp;
    
    R >> r;  // Radius
    r = pow(r,RadiusExp);
    for (i = 0; i < P->GoalState.dim(); i++) {
      zx[i] = (P->UpperState[i] - P->LowerState[i])*
	sqrt(P->GoalState.dim())*r*zx[i]/w + 
	P->GoalState[i];
    }
    
    // Check if sample is within bounds
    success = true;
    for (i = 0; i < P->GoalState.dim(); i++) {
      if ((zx[i] >= P->UpperState[i])||(zx[i] <= P->LowerState[i]))
	success = false;
    }
  }

  return zx;
}



vector RRTPolar::SelectBestInput(const vector &x1, const vector &x2, 
				 vector &nx_best, bool &success)
{
  vector u,u_best,nx;
  double d,dg,dmax,d_min;
  success = false;
  d_min = INFINITY;
  dg  = P->Metric(x1,x2);
  dmax  = P->Metric(P->LowerState,P->UpperState);
  list<vector> il = P->GetInputs(x1);
  forall(u,il) {
    //nx = P->Integrate(x1,u,PlannerDeltaT*sqrt(dg/dmax));  // Slow down near goal
    nx = P->Integrate(x1,u,PlannerDeltaT);
    d  = P->Metric(nx,x2);
    if ((d < d_min)&&(P->Satisfied(nx))&&(x1 != nx)) 
      {
	d_min = d; u_best = u; nx_best = nx; success = true;
      }
  }
  
  //cout << "u_best: " << u_best << "\n");

  return u_best;
}



// *********************************************************************
// *********************************************************************
// CLASS:     RRTHull
// 
// A simple exploration of what happens to the RRT in a large disc.
// Only works for 2DPoint model!!!!!
// *********************************************************************
// *********************************************************************

RRTHull::RRTHull(Problem *p):RRT(p) {
  Radius = 100000000.0;
}



vector RRTHull::ChooseState() {
  double theta;
  vector v(2);

  R >> theta;  theta *= 2.0*PI;

  v[0] = Radius*cos(theta);
  v[1] = Radius*sin(theta);

  return v;
}



// *********************************************************************
// *********************************************************************
// CLASS:     RRTExtCon
// 
// The planner presented in Kuffner, LaValle, 2000.  There are two
// trees, and the attempt to connect them is greedy.
// *********************************************************************
// *********************************************************************

RRTExtCon::RRTExtCon(Problem *p):RRTDual(p) {
}



bool RRTExtCon::Plan()
{
  int i;
  vector nx,nx_prev;
  node nn,nn2;
  bool connected;
  list<node> path1;

  connected = false;

  float t = used_time();

  //G.clear();
  if (G.number_of_nodes() == 0) {
    nn = G.new_node(P->InitialState);
    GEdgeTime[nn] = 0.0;
  }
  //G2.clear();
  if (G2.number_of_nodes() == 0) {
    nn = G2.new_node(P->GoalState);
    G2EdgeTime[nn] = 0.0;
  }

  nn = G.first_node();
  nn2 = G2.first_node();

  i = 0;
  connected = false;
  while ((i < NumNodes) && (!connected)) {
    if (Extend(ChooseState(),G,nn)) { 
      // Update the goal RRT
      if (Connect(G.inf(nn),G2,nn2,false)) {
	if (GapSatisfied(G.inf(nn),G2.inf(nn2))) {
	  cout << "CONNECTED!!  Nodes: " << G.number_of_nodes()+G2.number_of_nodes()
	       << "\n";
	  connected = true;
	  RecoverSolution(nn,nn2);
	}
      }
    }
    i++;

    if ((!connected)&&(Extend(ChooseState(),G2,nn,false))) { 
      // Update the initial RRT
      if (Connect(G2.inf(nn),G,nn2)) { 
	if (GapSatisfied(G.inf(nn),G2.inf(nn2))) {
	  cout << "CONNECTED!!  Nodes: " << G.number_of_nodes()+G2.number_of_nodes()
	       << "\n";
	  connected = true;
	  RecoverSolution(nn2,nn);
	}
      }
    }
    i++;
  }

  if (!connected)
    cout << "Failure to connect after " << G.number_of_nodes()+G2.number_of_nodes() 
	 << " nodes\n";

  cout << "Collision Detection Calls: " << SatisfiedCount << "\n";

  cout << "Planning Time: " << ((double)used_time(t)) << "s\n"; 

  return connected;
}



// *********************************************************************
// *********************************************************************
// CLASS:     RRTConCon
// 
// One the best dual-tree planners.  Be greedy in growing the trees
// and in trying to connect them to each other.  This is usually the
// fastest planner, but there is often a price for being greedy...
// *********************************************************************
// *********************************************************************

RRTConCon::RRTConCon(Problem *p):RRTDual(p) {
}



bool RRTConCon::Plan()
{
  int i;
  vector nx,nx_prev;
  node nn,nn2;
  bool connected;

  connected = false;

  // Code for time trials
  float t = used_time();

  //int total = 0;
  //for (int j=0; j < 100; j++) {

  //G.clear();
  if (G.number_of_nodes() == 0) {
    nn = G.new_node(P->InitialState);
    GEdgeTime[nn] = 0.0;
  }
  //G2.clear();
  if (G2.number_of_nodes() == 0) {
    nn = G2.new_node(P->GoalState);
    G2EdgeTime[nn] = 0.0;
  }

  nn = G.first_node();
  nn2 = G2.first_node();

  i = 0;
  connected = false;
  while ((i < NumNodes) && (!connected)) {
    if (Connect(ChooseState(),G,nn)) { 
      // Update the goal RRT
      if (Connect(G.inf(nn),G2,nn2,false)) {
	if (GapSatisfied(G.inf(nn),G2.inf(nn2))) {
	  cout << "CONNECTED!!  Nodes: " << 
	    G.number_of_nodes() + G2.number_of_nodes() 
	       << "\n";
	  connected = true;
	  RecoverSolution(nn,nn2); // Defined in RRTDual
	}
      }
    }
    i++;

    if ((!connected)&&(Connect(ChooseState(),G2,nn,false))) { 
      // Update the initial RRT
      if (Connect(G2.inf(nn),G,nn2)) { 
	if (GapSatisfied(G.inf(nn),G2.inf(nn2))) {
	  cout << "CONNECTED!!  Nodes: " << 
	    G.number_of_nodes() + G2.number_of_nodes()  
	       << "\n";
	  connected = true;
	  RecoverSolution(nn2,nn);
	}
      }
    }
    i++;
  }

  if (!connected)
    cout << "Failure to connect after " << 
      G.number_of_nodes()+G2.number_of_nodes() 
	 << " nodes\n";

  cout << "Collision Detection Calls: " << SatisfiedCount << "\n";

  // Code for time trials
  //total += i;
  //  }
  //cout << "Avg: " << total/100.0 << "\n";
  //cout << "Avg time: " << ((double)used_time(t)/100.0) << "\n"; 
  
  cout << "Planning Time: " << ((double)used_time(t)) << "s\n"; 

  return connected;
}


// *********************************************************************
// *********************************************************************
// CLASS:     RRTStar
// 
// An attempt to blend A^* ideas with the RRT.
// *********************************************************************
// *********************************************************************

RRTStar::RRTStar(Problem *p):RRT(p) {
  NodeDepth.init(G);
  OptimalityFactor = 0.9;
}



node RRTStar::NearestNeighbor(const vector &x, const GRAPH<vector,vector> &g,
			      bool forward = true) {
  double d,d_min;
  node n,n_best;
  
  n_best = g.choose_node(); // Keeps the warnings away
  d_min = INFINITY;
  forall_nodes(n,g) {
    d = P->Metric(g[n],x) +
      OptimalityFactor*P->Metric(P->InitialState,g[n]);
    //OptimalityFactor*PlannerDeltaT*NodeDepth[n]; // Here is the new part
    // The speed from Model used to appear in the second product above
    if (d < d_min) {
      d_min = d; n_best = n; 
    }
  } 

  return n_best;
}



bool RRTStar::Extend(const vector &x, 
		 GRAPH<vector,vector> &g,
		 node &nn, bool forward = true) {
  node n_best;
  vector nx,u_best;
  bool success;
  
  n_best = NearestNeighbor(x,g,forward);
  u_best = SelectBestInput(g.inf(n_best),x,nx,success); // nx gets next state
  if (success) {   // If a collision-free input was found
    nn = g.new_node(nx); // Make a new node with state nx
    // This NodeDepth relies on root initialization to zero!!!!
    NodeDepth[nn] = NodeDepth[n_best] + 1;
    //cout << "Node Depth: " << NodeDepth[nn] << "\n";
    g.new_edge(n_best,nn,u_best);
    //cout << "New node: " << g.inf(nn) << "   " << "u:" << u_best << "\n";
  }

  return success;
}



// *********************************************************************
// *********************************************************************
// CLASS:     PlannerHsu
// 
// The planner of Hsu, Latombe, Motwani (not an RRT).
// *********************************************************************
// *********************************************************************

PlannerHsu::PlannerHsu(Problem *p):RRT(p) {

  NeighborFactor = 0.02;
  NeighborRadius = 2.0;
  InverseWeights = node_map<double>(G);
  WeightScale = 0.0;

  // Set the step size
  StepSize = P->Metric(P->InitialState,P->Integrate(P->InitialState,
 	     P->GetInputs(P->InitialState).head(),PlannerDeltaT));

}


bool PlannerHsu::Extend(const vector &x, 
		 GRAPH<vector,vector> &g,
		 node &nn, bool forward = true) {

  bool success;
  vector u_best = P->GetInputs(x).head();
  vector x1,x2;
  node n_best;

  // Fix the graph for the initial case
  if ((g.number_of_nodes() == 1) && (InverseWeights[g.first_node()] == 0.0)) {
    InverseWeights[g.first_node()] = 1.0;
    WeightScale = 1.0;
  }

  // Select node
  n_best = SelectNode(g);
  //n_best = g.choose_node();
  x1 = g.inf(n_best);

  // Make random neighbor
  x2 = ChooseNearbyState(x1);

  // Attempt connection
  success = Connect(x1,x2);

  // nx gets next state
  if (success) {   // If a collision-free input was found
    nn = g.new_node(x2); // Make a new node with state nx
    UpdateWeights(nn,g);
    g.new_edge(n_best,nn,u_best);
    //cout << "New node: " << g.inf(nn) << "   " << "u:" << u_best << "\n";
  }

  return success;
}



bool PlannerHsu::Connect(const vector &x1, const vector &x2) {
  double a;
  int k;
  vector x;

  k = (int) (P->Metric(x1,x2) / StepSize); 

  for (a = 1.0/(k+1); a <= 1.0; a += 1.0/(k+1) ) {
    x = P->InterpolateState(x1,x2,a);
    if (!P->Satisfied(x))
      return false;
  }

  return true;
}



node PlannerHsu::SelectNode(const GRAPH<vector,vector> &g) {
  node n;
  double r,ptr,w;

  //cout << "W: ";
  //forall_nodes(n,g) {
  //  cout << InverseWeights[n] << " ";
  //}
  //cout << "\n";

  // Select the node
  R >> r; r *= WeightScale;
  ptr = 0.0;
  forall_nodes(n,g) {
    w = InverseWeights[n];
    ptr += w;
    if (ptr > r) {
      return n; 
    }
  }

  cout << "ERROR: No node was selected\n";
  return n;  // Just a fake to avoid warnings
}



vector PlannerHsu::ChooseNearbyState(const vector &x) {
  int i;
  double r,txi;
  vector rx;

  rx = x;  // Set the proper dimension
  for (i = 0; i < P->StateDim; i++) {
      txi = INFINITY;
      while ((txi < P->LowerState[i])||(txi > P->UpperState[i])) {
	R >> r; r -= 0.5;
	txi = x[i] + 
	  r * (P->UpperState[i] - P->LowerState[i]) * NeighborFactor;
      }
      rx[i] = txi;
  }

  return rx;
}



void PlannerHsu::UpdateWeights(const node &nn,
			   const GRAPH<vector,vector> &g) {
  vector x;
  node n;
  
  x = g.inf(nn);

  InverseWeights[nn] = 1.0;  // Count itself
  WeightScale += 1.0;
  forall_nodes(n,g) {
    if ((P->Metric(x,g.inf(n)) < NeighborRadius)&&
	(n != nn))
    {
      // Update the new node
      InverseWeights[nn] = 1/((1/InverseWeights[nn]) + 1);
      WeightScale += InverseWeights[nn] - 1/((1/InverseWeights[nn]) - 1);
      // Update the other nodes 
      InverseWeights[n] = 1/((1/InverseWeights[n]) + 1);
      WeightScale += InverseWeights[n] - 1/((1/InverseWeights[n]) - 1);
    }
  }
}



// *********************************************************************
// *********************************************************************
// CLASS:     RandomTree
// 
// Technically not an RRT.  Pick a vertex and input at random.
// *********************************************************************
// *********************************************************************

RandomTree::RandomTree(Problem *p):RRT(p) {
}


node RandomTree::NearestNeighbor(const vector &x, const GRAPH<vector,vector> &g,
			  bool forward = true) {

  return g.choose_node();
}


vector RandomTree::SelectBestInput(const vector &x1, const vector &x2, 
				 vector &nx_best, bool &success,
				 bool forward = true) {
  double r;
  list<vector> il;
  int k;
  vector u_best;

  il = P->GetInputs(x1);
  R >> r;

  k = (int) (r * il.length());

  u_best = il.inf(il[k]);

  nx_best = P->Integrate(x1,u_best,PlannerDeltaT);
  SatisfiedCount++;

  return il.inf(il[k]);
  
}

#include <LEDA/UNDEFINE_NAMES.h>






