//----------------------------------------------------------------------
//               The Motion Strategy Library (MSL)
//----------------------------------------------------------------------
//
// Copyright (c) 1998-2000 Iowa State University and Steve LaValle.  
// All Rights Reserved.
// 
// Permission to use, copy, and distribute this software and its 
// documentation is hereby granted free of charge, provided that 
// (1) it is not a component of a commercial product, and 
// (2) this notice appears in all copies of the software and
//     related documentation. 
// 
// Iowa State University and the author make no representations
// about the suitability or fitness of this software for any purpose.  
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------

#include <fstream.h>
#include <math.h>

#include "scene.h"
#include "defs.h"

#include <LEDA/REDEFINE_NAMES.h>

// Constructor
Scene::Scene(Problem *problem, string path = "") {

  SetProblem(problem);

  if ((path.length() > 0)&&(path[path.length()-1] != '/'))
    path += "/";

  FilePath = path;

  NumBodies = P->NumBodies;

  if (is_file(FilePath+"LowerWorld")) {
    file_istream fin(FilePath + "LowerWorld");
    fin >> LowerWorld;
    fin.close();
  }
  else {
    LowerWorld = vector(-50.0,-50.0,-50.0);
  }

  if (is_file(FilePath+"UpperWorld")) {
    file_istream fin2(FilePath + "UpperWorld");
    fin2 >> UpperWorld;
    fin2.close();
  }
  else {
    UpperWorld = vector(50.0,50.0,50.0);
  }

  if (is_file(FilePath+"GlobalCameraPosition")) {
    file_istream fin3(FilePath + "GlobalCameraPosition");
    fin3 >> GlobalCameraPosition;
    fin3.close();
  }
  else {
    GlobalCameraPosition = vector(3);

    GlobalCameraPosition[0] = (UpperWorld[0] + LowerWorld[0])/2.0;  
    GlobalCameraPosition[1] = (UpperWorld[1] + LowerWorld[1])/2.0;  
    GlobalCameraPosition[2] = UpperWorld[2] + (UpperWorld[1]-LowerWorld[1])/tan(38.0/180.0*PI);
    //    GlobalCameraPosition = vector(0.0,0.0,200.0);
  }

  if (is_file(FilePath+"GlobalCameraDirection")) {
    file_istream fin4(FilePath + "GlobalCameraDirection");
    fin4 >> GlobalCameraDirection;
    GlobalCameraDirection = GlobalCameraDirection.norm();
    fin4.close();
  }
  else {
    GlobalCameraDirection = vector(0.0,0.0,-1.0);
  }

  if (is_file(FilePath+"GlobalCameraZenith")) {
    file_istream fin4(FilePath + "GlobalCameraZenith");
    fin4 >> GlobalCameraZenith;
    GlobalCameraZenith = GlobalCameraZenith.norm();
    fin4.close();
  }
  else {
    GlobalCameraZenith = vector(0.0,1.0,0.0);
  }


  if (is_file(FilePath+"AttachedCameraPosition")) {
    file_istream fin3(FilePath + "AttachedCameraPosition");
    fin3 >> AttachedCameraPosition;
    fin3.close();
  }
  else {
    AttachedCameraPosition = vector(0.0,0.0,6.0);
  }

  if (is_file(FilePath+"AttachedCameraDirection")) {
    file_istream fin4(FilePath + "AttachedCameraDirection");
    fin4 >> AttachedCameraDirection;
    AttachedCameraDirection = AttachedCameraDirection.norm();
    fin4.close();
  }
  else {
    AttachedCameraDirection = vector(1.0,0.0,0.0);
  }

  if (is_file(FilePath+"AttachedCameraZenith")) {
    file_istream fin4(FilePath + "AttachedCameraZenith");
    fin4 >> AttachedCameraZenith;
    AttachedCameraZenith = AttachedCameraZenith.norm();
    fin4.close();
  }
  else {
    AttachedCameraZenith = vector(0.0,0.0,1.0);
  }

  if (is_file(FilePath+"AttachedCameraBody")) {
    file_istream fin4(FilePath + "AttachedCameraBody");
    fin4 >> AttachedCameraBody;
    fin4.close();
  }
  else {
    AttachedCameraBody = 0;
  }

}



void Scene::SetProblem(Problem *problem) {
  P = problem;
  NumBodies = P->NumBodies;
  GeomDim = P->GeomDim;
  SceneConfigurationDim = NumBodies * 6;  // 6 DOF for each body
}


// By default, don't change anything
vector Scene::StateToSceneConfiguration(const vector &x) {
  vector sc,q;
  int i;
  if (GeomDim == 3)
    return P->StateToConfiguration(x);
  else { // GeomDim == 2
    q = P->StateToConfiguration(x); // Three parameters per body
    // Blow out the vector to make 6 parameters per body
    sc = vector(NumBodies*6);
    for (i = 0; i < NumBodies; i++) {
      sc[6*i] = q[3*i];
      sc[6*i+1] = q[3*i+1];
      sc[6*i+2] = 0.0;
      sc[6*i+3] = 0.0;
      sc[6*i+4] = 0.0;
      sc[6*i+5] = q[3*i+2];
    }
    return sc;
  }
}


vector Scene::InterpolatedSceneConfiguration(const vector &x1, 
					     const vector &x2, 
					     const double &a) {

  return StateToSceneConfiguration(P->InterpolateState(x1,x2,a));
}

#include <LEDA/UNDEFINE_NAMES.h>
